"""
Python port of Perl script binarize_tags.pl

Binarizes unary XML milestone elements by assuming each milestone element begin a span stretching until the
next occurrence of the same element, or the end of the file
"""

from argparse import ArgumentParser
import io, re
from collections import defaultdict

def binarize(in_data):
	rev_lines = in_data.split("\n")[::-1]
	attrs = {}
	unary = defaultdict(int)
	edited_lines = []
	for line in rev_lines:
		elements = re.findall(r'<([^> ]+)([^>]*)/>',line)
		for tag, ats in elements:
			attrs[tag] = ats
			unary[tag] += 1
			line = re.sub(r'<' + tag + '[^>]*/>',r'</'+tag+'><'+tag+attrs[tag]+">",line)
		edited_lines.append(line)

	count = defaultdict(int)
	outlines = []
	for line in edited_lines[::-1]:
		tags = re.findall(r'</([^> ]+)>',line)
		for tag in tags:
			if tag in unary:
				count[tag]+=1

			if count[tag] == 1 and tag in unary:  # Destroy superfluous first closing tag
				line = line.replace("</"+tag+">","",1)
		outlines.append(line)

	finals = []
	for tag in sorted(unary.keys()):  # Final closing tags
		finals.append("</" + tag + ">")

	return "\n".join(outlines) + "\n".join(finals) + "\n"


if __name__=="__main__":
	parser = ArgumentParser()
	parser.add_argument("infile",action="store",help="file to process")

	opts = parser.parse_args()

	in_data = io.open(opts.infile,encoding="utf8").read().replace("\r","")
	binarized = binarize(in_data)
	print(binarized)




#!/usr/bin/python
# -*- coding: utf-8 -*-
import os, requests, io, shutil, sys

script_dir = os.path.dirname(os.path.realpath(__file__))

def download_file(url, local_path):
	try:
		sys.stderr.write("o Downloading model from " + url + "...\n")
		with requests.get(url, stream=True) as r:
			with io.open(local_path, 'wb') as f:
				shutil.copyfileobj(r.raw, f)
		sys.stderr.write("o Download successful\n")
	except Exception as e:
		sys.stderr.write("\n! Could not download model from " + url + "\n")
		sys.stderr.write(str(e))


def check_models(path=None):
# Check for missing models
	if path is None:
		model = "eng_flair_nner_distilbert.pt"
		path = os.sep.join([script_dir,"..","models","_sequence_taggers",model])
	else:
		model = path.split(os.sep)[-1]

	if not os.path.exists(path):
		server = "corpling.uis.georgetown.edu"
		resource = "/".join([server, "amir", "download", model])
		download_file("https://" + resource, os.sep.join([script_dir,"..","models","_sequence_taggers",model]))
	else:
		sys.stderr.write("o Model " + model + " already found in models/_sequence_taggers/")

if __name__ == "__main__":
	check_models()



"""
Script to read Scriptorium style SGML
into plain TT SGML: norm -> tok etc.
"""

import sys, io, re


def harvest_tt(text, keep_sgml=True):

	norm, pos, lemma = "","",""
	output = ""

	for line in text.split("\n"):
		m = re.search(' norm="([^"]+)"',line)
		if m is not None:
			norm = m.group(1)
		m = re.search(' lemma="([^"]+)"',line)
		if m is not None:
			lemma = m.group(1)
		m = re.search(' pos="([^"]+)"',line)
		if m is not None:
			pos = m.group(1)

		if keep_sgml:
			if "<" in line and re.search(r'(norm|pos|lemma)=|</(norm|pos|lemma)>',line) is None:
				output += line + "\n"

		if "</norm>" in line:
			if norm !="" and norm != " ":
				output += "\t".join([norm,pos,lemma])+"\n"

	return output


if __name__ == "__main__":

	if sys.argv[1] == "-g":
		from glob import glob
		files = glob(sys.argv[2])
		all_out = ""
		for file_ in files:
			input_text = io.open(file_, encoding="utf8").read().replace("\r", "")
			out_ = harvest_tt(input_text,keep_sgml=False)
			all_out += out_
		sys.stdout.buffer.write((all_out + "\n").encode('utf8'))

	else:

		input_text = io.open(sys.argv[1],encoding="utf8").read().replace("\r","")

		out_ = harvest_tt(input_text)
		sys.stdout.buffer.write((out_ + "\n").encode('utf8'))




#!/usr/bin/python
# -*- coding: utf-8 -*-

########################################################################
#                                                                      #
#  tokenization script for tagger preprocessing                        #
#  Adapted from TreeTagger tokenizer in Perl by:                       #
#          Helmut Schmid, IMS, University of Stuttgart                 #
#          Serge Sharoff, University of Leeds                          #
#  Description:                                                        #
#  - splits input text into tokens (one token per line)                #
#  - cuts off punctuation, parentheses etc.                            #
#  - disambiguates periods                                             #
#  - preserves SGML markup                                             #
#                                                                      #
#  - Ported to Python and adapted to Hebrew by Amir Zeldes             #
########################################################################


from argparse import ArgumentParser
import io, sys, re

PY3 = sys.version_info[0] > 2

# characters which have to be cut off at the beginning of a word
PChar = r"""\[¿¡{(`"‚„†‡‹‘’“”•'–—›«「"""

# characters which have to be cut off at the end of a word
FChar = r"""'\]}'`"),;:!?%‚„…†‡‰‹‘’“”•–—›'»」"""

# character sequences which have to be cut off at the beginning of a word
PClitic = ""

# character sequences which have to be cut off at the end of a word
FClitic = ""


def tokenize(text, abbr=None, add_sents=False):

	output = ""
	if add_sents:
		lines = []
		raw_lines = text.split("\n")
		for line in raw_lines:
			if len(line.strip())>0:
				lines.append("<s>" + line + "</s>")
	else:
		lines = text.split("\n")

	# Read the list of abbreviations and words
	Token = set([])
	if abbr is not None:
		abbr_lines = io.open(abbr,encoding="utf8").read().replace("\r","").strip().split("\n")
		for line in abbr_lines:
			Token.add(line)

	for line in lines:
		# replace newlines and tab characters with blanks
		line = line.replace("\t"," ").replace("\n"," ")

		# replace blanks within SGML tags
		sep1 = "□"
		sep2 = "■"

		if not PY3:
			sep1 = sep1.decode("utf8")
			sep2 = sep2.decode("utf8")

		find_tag_space = r'(<[^<> ]*) ([^<>]*>)'
		while re.search(find_tag_space,line) is not None:
			line = re.sub(find_tag_space,'\1'+sep1+'\2',line)

		# replace whitespace with a special character
		line = line.replace(" ",sep2)

		# restore SGML tags
		line = line.replace(sep1," ")
		line = line.replace(sep2,sep1)

		# prepare SGML-Tags for tokenization

		line = re.sub(r'(<[^<>]*>)',sep1 + "\\1" + sep1,line)
		line = re.sub(r'^' + sep1,"",line)
		line = re.sub(sep1 + r'$',"",line)
		line = re.sub(sep1*3 +"*",sep1,line)

		units = line.split(sep1)
		for i, unit in enumerate(units):
			if re.match(r"<.*>$",unit) is not None:
				# SGML tag
				output += unit + "\n"
			else:
				#add a blank at the beginning and the end of each segment
				email = r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
				url = r'https?://(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&/=]*)'
				acro = r"([A-Z]\.([A-Z]\.)+)"
				short_url = r"[-a-zA-Z0-9]{2,256}\.(com|org|edu|co|io|net)"

				# insert missing blanks after punctuation if not an abbreviation
				if unit not in Token and re.search(email,unit) is None and re.search(url,unit) is None \
					and re.search(acro, unit) is None and re.search(short_url, unit) is None:
					unit = " " + unit + " "
					unit = re.sub(r'\.\.\.'," ... ",unit)
					unit = re.sub(r'([;!?])([^ ])',r'\1 \2', unit)
					unit = re.sub(r'([.,:])([^ 0-9.])', r'\1 \2',unit)
				else:
					unit = " " + unit + " "

				subunits = unit.split()
				for subunit in subunits:
					suffix=""

					# separate punctuation and parentheses from words
					while True:
						finished = True
						# cut off preceding punctuation
						m = re.match(r'(['+PChar+'])(.)',subunit)
						if m is not None:
							m1 = m.group(1)
							subunit = re.sub('^(['+PChar+'])(.)',r'\2',subunit)
							output += m1 + "\n"
							finished = 0
						# cut off trailing punctuation
						m = re.search(r'(.)(['+FChar+'])$',subunit)
						if m is not None:
							m2 = m.group(2)
							subunit = re.sub(r'(.)(['+FChar+'])$',r'\1',subunit)
							suffix = m2 + "\n" + suffix
							finished = 0

						# cut off trailing periods if punctuation precedes
						m = re.search(r'(['+FChar+'])\.$',subunit)
						if m is not None:
							subunit = re.sub(r'(['+FChar+'])\.$','',subunit,count=1)
							suffix = ".\n" + suffix
							if subunit == "":
								subunit = m.group(1)
							else:
								suffix = m.group(1) + "\n" + suffix
							finished = False
						if finished:
							break

					# handle explicitly listed tokens
					if subunit in Token:
						output += subunit + "\n" + suffix
						continue

					# abbreviations of the form A. or U.S.A.
					if re.match(r'([A-Za-z-]\.)+$',subunit) is not None:
						output += subunit + "\n" + suffix
						continue

					# e-mail addresses
					if re.search(email,subunit) is not None or re.search(url,subunit) is not None:
						output += subunit + "\n" + suffix
						continue

					# disambiguate periods
					m = re.match(r'(..*)\.$',subunit)
					if m is not None and subunit != "...":
						subunit = m.group(1)
						suffix = ".\n" + suffix
						if subunit in Token:
							output += subunit + "\n" + suffix
							continue

					# cut off clitics
					while re.match(r'(--)(.)',subunit) is not None:
						m = re.match(r'(--)(.)',subunit)
						subunit = re.sub(r'(--)(.)',r'\2',subunit)
						output += m.group(1) + "\n"

					if PClitic != '':
						while re.match(r'('+PClitic+')(.)',subunit) is not None:
							m = re.match(r'('+PClitic+')(.)',subunit)
							subunit = re.sub(r'('+PClitic+')(.)',r'\2',subunit)
							output += m.group(1) + "\n"

					while re.search(r'(.)(--)$',subunit) is not None:
						m = re.search(r'(.)(--)$',subunit)
						subunit = re.sub(r'(.)(--)$',r'\1',subunit)
						suffix = m.group(2) + "\n" + suffix

					if FClitic != "":
						while re.search(r'(.)('+FClitic+')$',subunit) is not None:
							m = re.search(r'(.)('+FClitic+')$',subunit)
							subunit = re.sub(r'(.)('+FClitic+')$',r"\1")
							suffix = m.group(2) + "\n" + suffix
					output+=subunit + "\n" + suffix
	return output


if __name__ == "__main__":

	p = ArgumentParser()
	p.add_argument("infile")
	p.add_argument("-a","--abbreviations",action="store",default=None,help="File name for list of abbreviations and other tokens to leave alone")

	opts = p.parse_args()

	input_text = io.open(opts.infile,encoding="utf8").read().replace("\r","").strip()
	output = tokenize(input_text,opts.abbreviations)
	print(output)



"""
modules/xrenner_preprocess.py

Prepare parser output for entity and coreference resolution

Author: Amir Zeldes
"""

from .xrenner_marker import lookup_has_entity

def add_negated_parents(conll_tokens, offset):
	"""
	Sets the neg_parent property on tokens whose head dominates a negation

	:param conll_tokens: token list for this document
	:param offset: token ID reached in last sentence
	:return: None
	"""

	for token in conll_tokens[offset:]:
		parent_id = token.head
		if parent_id != "0":
			if conll_tokens[int(parent_id)].negated:
				token.neg_parent = True


def add_child_info(conll_tokens, child_funcs, child_strings, lex):
	"""
	Adds a list of all dependent functions and token strings to each parent token
	
	:param conll_tokens: The ParsedToken list so far
	:param child_funcs: Dictionary from ids to child functions
	:param child_strings: Dictionary from ids to child strings
	:return: void
	"""
	for child_id in child_funcs:
		if child_id > len(conll_tokens)-1:
			continue
		for func in child_funcs[child_id]:
			if func not in conll_tokens[child_id].child_funcs:
				conll_tokens[child_id].child_funcs.append(func)
				if lex.filters["neg_func"].match(func):
					conll_tokens[child_id].negated = True
		for tok_text in child_strings[child_id]:
			if tok_text not in conll_tokens[child_id].child_strings:
				conll_tokens[child_id].child_strings.append(tok_text)


def postprocess_parser(conll_tokens, tokoffset, children, stop_ids, lex):
	for tok1 in conll_tokens[tokoffset + 1:]:
		if tok1.text == "-LSB-" or tok1.text == "-RSB-":
			tok1.pos = tok1.text
			tok1.func = "punct"
			tok1.head = "0"
		if lex.filters["mark_head_pos"].match(tok1.pos) is not None:
			entity_candidate = tok1.text + " "
			for tok2 in conll_tokens[int(tok1.id) + 1:]:
				if lex.filters["mark_head_pos"].match(tok2.pos) is not None:
					entity_candidate += tok2.text + " "
					### DEBUG BREAKPOINT ###
					if entity_candidate.strip() == lex.debug["ana"]:
						pass
					if entity_candidate.strip() in lex.entities:  # Entity matched, check if all tokens are inter-connected
						for tok3 in conll_tokens[int(tok1.id):int(tok2.id)]:
							# Ensure right most token has head outside entity:
							if int(tok2.head) > int(tok2.id) or int(tok2.head) < int(tok1.id):
								if (int(tok3.head) < int(tok1.id) or int(tok3.head) > int(tok2.id)) and tok3.id in children[tok3.head]:
									children[tok3.head].remove(tok3.id)
									tok3.head = tok2.id
									children[tok3.head].append(tok3.id)
									break
				else:
					break
		# Check for apposition pointing back to immediately preceding proper noun token -
		# typical (German model) MaltParser name behavior
		if lex.filters["apposition_func"].match(tok1.func) is not None and not tok1.id == "1":
			if lex.filters["proper_pos"].match(conll_tokens[int(tok1.id) - 1].pos) is not None and conll_tokens[
						int(tok1.id) - 1].id == tok1.head:
				tok1.func = "xrenner_fix"
				children[str(int(tok1.id) - 1)].append(tok1.id)
				stop_ids[tok1.id] = True

		# Check for [city], [state/country] apposition -
		# typical (English model) Stanford parser behavior
		if tok1.text == lex.debug["ana"]:
			a=5
		if lex.filters["apposition_func"].match(tok1.func) is not None and not int(tok1.id) < 3:
			if conll_tokens[int(tok1.id) - 1].text.strip() == ",":
				tok_minus2 = conll_tokens[int(tok1.id) - 2]
				tok1_head = conll_tokens[int(tok1.head)]
				if lex.filters["proper_pos"].match(tok_minus2.pos) is not None:
					if (tok_minus2.id == tok1.head and (lookup_has_entity(tok1.text, tok1.lemma, "place", lex) and not lookup_has_entity(tok_minus2.text, tok_minus2.lemma, "place", lex) or \
						lookup_has_entity(tok_minus2.text, tok_minus2.lemma, "place", lex))) or \
						not lookup_has_entity(tok1_head.text, tok1_head.lemma, "place", lex) and lookup_has_entity(tok1.text, tok1.lemma, "place", lex):
							tok1.func = "xrenner_fix"
							if tok1.id not in children[tok_minus2.id]:
								if tok_minus2.head != tok1.id:  # Avoid creating a cycle
									children[tok_minus2.id].append(tok1.id)

		# Check for markable projecting beyond an apposition to itself and remove from children on violation
		if lex.filters["apposition_func"].match(tok1.func) is not None and not tok1.id == "1":
			for tok2 in conll_tokens[int(tok1.id) + 1:]:
				if tok2.head == tok1.head and lex.filters["non_link_func"].match(tok2.func) is None and tok2.id in children[tok2.head]:
					children[tok2.head].remove(tok2.id)


def replace_conj_func(conll_tokens, tokoffset, lex):
	"""
	Function to replace functions of tokens matching the conjunction function with their parent's function
	
	:param conll_tokens: The ParsedToken list so far
	:param tokoffset: The starting token for this sentence
	:param lex: the LexData object with gazetteer information and model settings
	:return: void
	"""

	for token in conll_tokens[tokoffset:]:
		## DEBUG POINT ##
		if token.text == lex.debug["ana"]:
			pass

		if lex.filters["conjunct_func"].match(token.func) is not None:
			for child_func in conll_tokens[int(token.head)].child_funcs:
				token.child_funcs.append(child_func)
			token.func = conll_tokens[int(token.head)].func
			token.head = conll_tokens[int(token.head)].head
			token.coordinate = True



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Main class file for Xrenner() class

Author: Amir Zeldes
"""

from collections import OrderedDict
from .xrenner_out import *
from .xrenner_classes import *
from .xrenner_coref import *
from .xrenner_preprocess import *
from .xrenner_marker import make_markable
from .xrenner_lex import *
from .xrenner_postprocess import postprocess_coref
from .depedit import DepEdit
import ntpath, os, io, decimal

decimal.getcontext().rounding = decimal.ROUND_DOWN

class Xrenner:

	def __init__(self, model="eng", override=None, rule_based=False, no_seq=False):
		"""
		Main class for xrenner coreferencer. Invokes the load method to read model data.
		
		:param model:  model directory in models/ specifying settings and gazetteers for this language (default: eng)
		:param override: name of a section in models/override.ini if configuration overrides should be applied
		:param rule_based: do not use machine learning classifiers for coreference resolution
		:param no_seq: do not use machine learning sequence taggers for entity resolution
		"""

		self.rule_based = rule_based
		self.no_seq = no_seq
		self.load(model, override)
		if "depedit.ini" in self.lex.model_files:
			depedit_config = self.lex.model_files["depedit.ini"]
			self.depedit = DepEdit(depedit_config, options=type('', (), {"kill":"supertoks", "quiet":True})())
		else:
			self.depedit = None
		self.token_count = 0
		self.sentence_count = 0

	def load(self, model="eng", override=None):
		"""
		Method to load model data. Normally invoked by constructor, but can be repeated to change models later.

		:param model:  model directory in models/ specifying settings and gazetteers for this language (default: eng)
		:param override: name of a section in models/override.ini if configuration overrides should be applied
		:return: None
		"""

		self.model = model
		self.override = override
		self.lex = LexData(self.model, self, self.override, self.rule_based, self.no_seq)

	def set_doc_name(self, name):
		"""
		Method to manually set the name of the document being processed, rather than deriving it from an input file name.

		:param name: string, the name to give the document
		:return: None
		"""

		self.docname = name
		self.lex.docname = name  # Copy in lex, in case we need access in nested object

	def check_model(self, path=None):
		# Check for large model files which should be in models/_sequence_taggers/
		from .get_models import check_models
		check_models(path)

	def analyze(self, infile, out_format):
		"""
		Method to run coreference analysis with loaded model
		
		:param infile: file name of the parse file in the conll10 format, or the pre-read parse itself
		:param format: format to determine output type, one of: html, paula, webanno, conll, onto, unittest
		:return: output based on requested format
		"""

		# Check if this is a file name from the main script or a parse delivered in an import or unittest scenario
		if "\t" in infile or isinstance(infile,list):  # This is a raw parse as string or list, not a file name
			self.docpath = os.path.dirname(os.path.abspath("."))
			if self.lex.docname is None:
				self.set_doc_name("untitled")
			if not isinstance(infile,list):
				infile = infile.replace("\r","").split("\n")
		else:  # This is a file name, extract document name and path, then read the file
			self.docpath = os.path.dirname(os.path.abspath(infile))
			self.set_doc_name(clean_filename(ntpath.basename(infile)))
			try:
				temp = io.open(infile, encoding="utf8")
				temp = temp.read()
			except UnicodeDecodeError:
				temp = io.open(infile, encoding="ISO 8859-1")
				temp = temp.read()
			infile = temp.replace("\r","").split("\n")

		# Empty cached lists of incompatible pairs
		self.lex.incompatible_mod_pairs = set([])
		self.lex.incompatible_isa_pairs = set([])


		if self.depedit is not None:
			infile = self.depedit.run_depedit(infile, self.docname)
		if not isinstance(infile,list):
			infile = infile.split("\n")
		# Count non-comment, non-empty lines to get token count
		self.token_count = len(list([tok for tok in infile if not (tok.startswith("#") or len(tok)==0)]))
		self.sentence_count = len(list([tok for tok in infile if tok.startswith("1\t")]))

		# Lists and dictionaries to hold tokens and markables
		self.conll_tokens = []
		self.conll_tokens.append(ParsedToken(0, "ROOT", "--", "XX", "", -1, "NONE", Sentence(1, 0, ""), [], [], [], self.lex))
		self.markables = []
		self.markables_by_head = OrderedDict()
		self.markstart_dict = defaultdict(list)
		self.markend_dict = defaultdict(list)

		self.tokoffset = 0
		self.sentlength = 0
		self.markcounter = 1
		self.groupcounter = 1

		self.children = defaultdict(list)
		self.descendants = {}
		self.child_funcs = defaultdict(list)
		self.child_strings = defaultdict(list)

		# Dereference object classes to method globals for convenience
		lex = self.lex
		conll_tokens = self.conll_tokens
		markstart_dict = self.markstart_dict
		markend_dict = self.markend_dict

		sentences = []
		self.sent_num = 1
		quoted = False
		current_sentence = Sentence(self.sent_num, self.tokoffset, "")

		lex.coref_rules = lex.non_speaker_rules

		seq_preds = None
		text = "\n".join(infile).strip()
		sents = text.split("\n\n")
		s_texts = []
		lemmas = []
		lemma_freqs = defaultdict(float)
		for s in sents:
			tablines = [line.split("\t") for line in s.split("\n") if "\t" in line]
			no_super = [line[1] for line in tablines if "-" not in line[0]]
			lemmas += [line[2] for line in tablines if "-" not in line[0]]
			s_texts.append(" ".join(no_super))

		if lex.sequencer is not None:  # Sequence label all tokens before reading sentences
			seq_preds = lex.sequencer.predict_proba(s_texts)

		for myline in infile:
			if "speaker" in myline and "=" in myline and myline.startswith("#"):  # speaker annotation
				current_sentence.speaker = myline.split("=")[1].strip()
				lex.coref_rules = lex.speaker_rules
			elif "s_type" in myline and "=" in myline and myline.startswith("#"):  # s_type annotation
				current_sentence.s_type = myline.split("=")[1].strip()
			elif myline.find("\t") > 0:  # Only process lines that contain tabs (i.e. conll tokens)
				current_sentence.token_count += 1
				cols = myline.split("\t")
				if "." in cols[0] or "-" in cols[0]:  # conllu multi-token line or decimal ID virtual token
					continue  # Not currently supported
				if lex.filters["open_quote"].match(cols[1]) is not None and quoted is False:
					quoted = True
				elif lex.filters["close_quote"].match(cols[1]) is not None and quoted is True:
					quoted = False
				if lex.filters["question_mark"].match(cols[1]) is not None:
					current_sentence.mood = "question"
				if cols[3] in lex.func_substitutes_forward and int(cols[6]) > int(cols[0]):
					tok_func = re.sub(lex.func_substitutes_forward[cols[3]][0],lex.func_substitutes_forward[cols[3]][1],cols[7])
				elif cols[3] in lex.func_substitutes_backward and int(cols[6]) < int(cols[0]):
					tok_func = re.sub(lex.func_substitutes_backward[cols[3]][0],lex.func_substitutes_backward[cols[3]][1],cols[7])
				else:
					tok_func = cols[7]
				head_id = "0" if cols[6] == "0" else str(int(cols[6]) + self.tokoffset)
				this_tok = ParsedToken(str(int(cols[0]) + self.tokoffset), cols[1], cols[2], cols[3], cols[5],
												head_id, tok_func, current_sentence, [], [], [], lex, quoted, cols[8], cols[9])
				if seq_preds is not None:
					this_tok.seq_pred = seq_preds[int(cols[0]) + self.tokoffset -1]
				conll_tokens.append(this_tok)
				self.sentlength += 1
				# Check not to add a child if this is a function which discontinues the markable span
				if not (lex.filters["non_link_func"].match(tok_func) is not None or lex.filters["non_link_tok"].match(cols[1]) is not None):
					if cols[6] != "0":  # Do not add children to the 'zero' token
						self.children[str(int(cols[6]) + self.tokoffset)].append(str(int(cols[0]) + self.tokoffset))
				self.child_funcs[(int(cols[6]) + self.tokoffset)].append(tok_func)
				self.child_strings[(int(cols[6]) + self.tokoffset)].append(cols[1])
			elif self.sentlength > 0:
				#self.process_sentence(self.tokoffset, current_sentence)
				self.sent_num += 1
				if self.sentlength > 0:
					self.tokoffset += self.sentlength
				current_sentence.length = self.sentlength
				sentences.append(current_sentence)
				current_sentence = Sentence(self.sent_num, self.tokoffset, "")

				self.sentlength = 0

		# Handle leftover sentence which did not have trailing newline
		if self.sentlength > 0:
			#self.process_sentence(self.tokoffset, current_sentence)
			current_sentence.length = self.sentlength
			sentences.append(current_sentence)

		# Get lemma frequencies for this document
		token_count = float(len(lemmas))
		lex.token_count = token_count
		for lemma in set(lemmas):
			###lemma_freqs[lemma] = float(round(decimal.Decimal(lemmas.count(lemma)/token_count),4))
			lemma_freqs[lemma] = lemmas.count(lemma)
		lex.lemma_freqs = lemma_freqs
		for tok in conll_tokens:
			tok.lemma_freq = lemma_freqs[tok.lemma]

		self.tokoffset = 0
		for snum, sentence in enumerate(sentences):
			sentence.text = s_texts[snum]
			self.tokoffset += sentence.start_offset - self.tokoffset
			self.process_sentence(sentence.start_offset,sentence)

		marks_to_add = []
		dump = "" ###
		if lex.filters["seek_verb_for_defs"]:
			for mark in self.markables:
				if mark.definiteness == "def" and mark.antecedent == "none" and mark.form == "common" and \
				(lex.filters["event_def_entity"] == mark.entity or lex.filters["abstract_def_entity"] == mark.entity):
					for tok in conll_tokens[0:mark.start]:
						if lex.filters["verb_head_pos"].match(tok.pos):
							dump += str(mark.start)+"-"+str(mark.end) + ";"+tok.id+"-"+tok.id+"\t"+str(lex.docname)+"\t"+tok.text+"\t"+mark.head.text+"\t"+str(mark.sent_num-tok.sentence.sent_num)+"\t"
							if stems_compatible(tok,mark.head,lex):
								comp = "T"
								dump += comp + "\n"
								v_antecedent = make_markable(tok,conll_tokens,{},tok.sentence.start_offset,tok.sentence,[],lex)
								mark.antecedent = v_antecedent
								mark.coref_type = "coref"
								v_antecedent.entity = mark.entity
								v_antecedent.subclass = mark.subclass
								v_antecedent.definiteness = "none"
								v_antecedent.form = "verbal"
								v_antecedent.infstat = "new"
								v_antecedent.group = mark.group
								v_antecedent.id = "referent_" + v_antecedent.head.id
								marks_to_add.append(v_antecedent)
							else:
								comp = "F"


		for mark in marks_to_add:
			markstart_dict[mark.start].append(mark)
			markend_dict[mark.end].append(mark)
			self.markables_by_head[mark.head.id] = mark
			self.markables.append(mark)


		postprocess_coref(self.markables, lex, markstart_dict, markend_dict, self.markables_by_head, conll_tokens)

		if out_format == "paula":
			try:
				self.serialize_output(out_format)
				return True
			except:
				return False
		else:
			return self.serialize_output(out_format, infile)

	def analyze_markable(self, mark, lex):
		"""
		Find entity, agreement and cardinality information for a markable
		:param mark: The :class:`.Markable` object to analyze
		:param lex: the :class:`.LexData` object with gazetteer information and model settings
		:return: void
		"""

		mark.text = mark.text.strip()
		mark.core_text = mark.core_text.strip()
		# DEBUG POINT
		if mark.text == lex.debug["ana"]:
			a=5
		tok = mark.head
		if lex.filters["proper_pos"].match(tok.pos) is not None:
			mark.form = "proper"
			mark.definiteness = "def"
		elif lex.filters["pronoun_pos"].match(tok.pos) is not None:
			mark.form = "pronoun"
			# Check for explicit indefinite morphology in morph feature of head token
			if "indef" in mark.head.morph.lower():
				mark.definiteness = "indef"
			else:
				mark.definiteness = "def"
		else:
			mark.form = "common"
			# Check for explicit definite morphology in morph feature of head token
			if "def" in mark.head.morph.lower() and "indef" not in mark.head.morph.lower():
				mark.definiteness = "def"
				# Chomp definite information not to interfere with agreement
				mark.head.morph = re.sub("def", "_", mark.head.morph)
			else:
				# Check if any children linked via a link function are definite markings
				children_are_def_articles = (lex.filters["definite_articles"].match(maybe_article) is not None for
											 maybe_article in
											 [mark.head.text, mark.text.split(" ")[0]] + mark.head.child_strings)
				children_are_possessors =  (lex.filters["definite_possessive_func"].match(func) is not None for func in mark.head.child_funcs)
				if any(children_are_def_articles) or any(children_are_possessors):
					mark.definiteness = "def"
				else:
					mark.definiteness = "indef"

		# Find agreement alternatives unless cardinality has set agreement explicitly already (e.g. to 'plural'/'dual' etc.)
		if mark.cardinality == 0 or mark.agree == '':
			mark.alt_agree = resolve_mark_agree(mark, lex)
		if mark.alt_agree is not None and mark.agree == '':
			mark.agree = mark.alt_agree[0]
		elif mark.alt_agree is None:
			mark.alt_agree = []
		if mark.agree != mark.head.morph and mark.head.morph != "_" and mark.head.morph != "--" and mark.agree != \
				lex.filters["aggregate_agree"]:
			mark.agree = mark.head.morph
			mark.agree_certainty = "mark_head_morph"
			mark.alt_agree.append(mark.head.morph)

		# cardinality resolve, only resolve here if it hasn't been set before (as in coordination markable)
		if mark.cardinality == 0:
			mark.cardinality = resolve_cardinality(mark, lex)

		if mark.agree in lex.filters["agree_entity_mapping"]:
			mark.entity = lex.filters["agree_entity_mapping"][mark.agree]
		else:
			resolve_mark_entity(mark, lex)

		if lex.entity_oracle is not None:
			if lex.oracle_counters is None:
				lex.oracle_counters = [0,0,0]
			sent_text = mark.sentence.text
			lex.oracle_counters[2] += 1
			if sent_text in lex.entity_oracle:
				m_start = mark.start - mark.sentence.start_offset
				m_end = mark.end - mark.sentence.start_offset
				if (m_start, m_end) in lex.entity_oracle[sent_text]:
					lex.oracle_counters[0] += 1
					if mark.entity != lex.entity_oracle[sent_text][(m_start, m_end)]:
						lex.oracle_counters[1] += 1
					mark.entity = lex.entity_oracle[sent_text][(m_start, m_end)]


		if "ablations" in lex.debug:
			if "no_subclasses" in lex.debug["ablations"]:
				mark.subclass = mark.entity
				mark.alt_subclasses = mark.alt_entities

	def serialize_output(self, out_format, parse=None):
		"""
		Return a string representation of the output in some format, or generate PAULA directory structure as output
		
		:param out_format: the format to generate, one of: html, paula, webanno, conll, onto, unittest
		:param parse: the original parse input fed to xrenner; only needed for unittest output
		:return: specified output format string, or void for paula
		"""
		conll_tokens = self.conll_tokens
		markables, markstart_dict, markend_dict = self.markables, self.markstart_dict, self.markend_dict

		if out_format == "html":
			rtl = True if self.model in ["heb","ara"] else False
			return output_HTML(conll_tokens, markstart_dict, markend_dict, rtl)
		elif out_format == "paula":
			output_PAULA(conll_tokens, markstart_dict, markend_dict, self.docname, self.docpath)
		elif out_format == "webanno":
			return output_webanno(conll_tokens[1:], markables)
		elif out_format == "webannotsv":
			return output_webannotsv(conll_tokens[1:], markables)
		elif out_format == "conll":
			return output_conll(conll_tokens, markstart_dict, markend_dict, self.docname, False)
		elif out_format == "conll_sent":
			return output_conll_sent(conll_tokens, markstart_dict, markend_dict, self.docname, True)
		elif out_format == "onto":
			return output_onto(conll_tokens, markstart_dict, markend_dict, self.docname)
		elif out_format == "unittest":
			from .xrenner_test import generate_test
			return generate_test(conll_tokens, markables, parse, self.model)
		elif out_format == "none":
			return ""
		else:
			return output_SGML(conll_tokens, markstart_dict, markend_dict)

	def process_sentence(self, tokoffset, sentence):
		"""
		Function to analyze a single sentence
		
		:param tokoffset: the offset in tokens for the beginning of the current sentence within all input tokens
		:param sentence: the Sentence object containing mood, speaker and other information about this sentence
		:return: void
		"""
		def is_eligible_submark_head(head_tok):
			# Note this function ignores pos_func_heads combos
			if lex.filters["mark_head_pos"].match(head_tok.pos) is not None:
				if lex.filters["mark_forbidden_func"].match(head_tok.func) is None:
					return True
			return False

		markables = self.markables
		markables_by_head = self.markables_by_head

		lex = self.lex
		conll_tokens = self.conll_tokens[:tokoffset+sentence.token_count+1]
		child_funcs = self.child_funcs
		child_strings = self.child_strings
		children = self.children
		descendants = self.descendants
		markstart_dict = self.markstart_dict
		markend_dict = self.markend_dict

		use_sequencer = True if lex.sequencer is not None else False

		# Add list of all dependent funcs and strings to each token
		add_child_info(conll_tokens, child_funcs, child_strings, lex)
		add_negated_parents(conll_tokens, tokoffset)

		mark_candidates_by_head = OrderedDict()
		stop_ids = {}
		for tok1 in conll_tokens[tokoffset + 1:]:
			stop_ids[tok1.id] = False  # Assume all tokens are head candidates
			tok1.sent_position = float(int(tok1.id) - tokoffset) / sentence.token_count # Add relative token positions at sentence as percentages
			tok1.doc_position = float(int(tok1.id)) / self.token_count # Add relative token positions at document as percentages
			tok1.head_text = conll_tokens[int(tok1.head)].text  # Save parent text for later dependency checks
			tok1.head_pos = conll_tokens[int(tok1.head)].pos  # Save parent POS for later dependency checks

		# Post-process parser input based on entity list if desired
		if lex.filters["postprocess_parser"]:
			postprocess_parser(conll_tokens, tokoffset, children, stop_ids, lex)

		# Revert conj token function to parent function
		replace_conj_func(conll_tokens, tokoffset, lex)

		# Enrich tokens with modifiers and parent head text
		for token in conll_tokens[tokoffset:]:
			for child in children[token.id]:
				if lex.filters["mod_func"].match(conll_tokens[int(child)].func) is not None:
					token.modifiers.append(conll_tokens[int(child)])
			token.head_text = conll_tokens[int(token.head)].text
			# Check for lexical possessives to dynamically enhance hasa information
			if lex.filters["possessive_func"].match(token.func) is not None:
				# Check that neither possessor nor possessed is a pronoun
				if lex.filters["pronoun_pos"].match(token.pos) is None and lex.filters["pronoun_pos"].match(conll_tokens[int(token.head)].pos) is None:
					lex.hasa[token.text][conll_tokens[int(token.head)].text] += 2  # Increase by 2: 1 for attestation, 1 for pertinence in this document
					lex.hasa[token.lemma][conll_tokens[int(token.head)].text] += 1
			# Check if func2 has additional possessor information
			if token.func2 != "_":
				if lex.filters["possessive_func"].match(token.func2) is not None:
					if lex.filters["pronoun_pos"].match(token.pos) is None and lex.filters["pronoun_pos"].match(conll_tokens[int(token.head2)+tokoffset].pos) is None:
						lex.hasa[token.text][conll_tokens[int(token.head2)+tokoffset].text] += 2  # Increase by 2: 1 for attestation, 1 for pertinence in this document
						lex.hasa[token.lemma][conll_tokens[int(token.head2)+tokoffset].text] += 1

		# Find dead areas
		for tok1 in conll_tokens[tokoffset + 1:]:
			# Affix tokens can't be markable heads - assume parser error and fix if desired
			# DEBUG POINT
			if tok1.text == lex.debug["ana"]:
				a=5
			if use_sequencer:
				if tok1.seq_pred[0] == "O" and tok1.seq_pred[1] > lex.filters["sequencer_nonref_thresh"] and lex.filters["sequencer_nonref_pos"].match(tok1.pos) is not None:
					if not any([lex.filters["sequencer_nonref_forbidden_childfunc"].match(f) is not None for f in tok1.child_funcs]):
						stop_ids[tok1.id] = True
			if lex.filters["postprocess_parser"]:
				if ((lex.filters["mark_head_pos"].match(tok1.pos) is not None and lex.filters["mark_forbidden_func"].match(tok1.func) is None) or
				pos_func_combo(tok1.pos, tok1.func, lex.filters["pos_func_heads"])) and not (stop_ids[tok1.id]):
					if tok1.text.strip() in lex.affix_tokens:
						stop_ids[tok1.id] = True
						for child_id in sorted(children[tok1.id], reverse=True):
							child = conll_tokens[int(child_id)]
							if ((lex.filters["mark_head_pos"].match(child.pos) is not None and lex.filters["mark_forbidden_func"].match(child.func) is None) or
							pos_func_combo(child.pos, child.func, lex.filters["pos_func_heads"])) and not (stop_ids[child.id]):
								child.head = tok1.head
								tok1.head = child.id
								# Make the new head be the head of all children of the affix token
								for child_id2 in children[tok1.id]:
									if not child_id2 == child_id:
										conll_tokens[int(child_id2)].head = child.id
										children[tok1.id].remove(child_id2)
										children[child.id].append(child_id2)

								# Assign the function of the affix head to the new head and vice versa
								temp_func = child.func
								child.func = tok1.func
								tok1.func = temp_func
								children[tok1.id].remove(child.id)
								children[child.id].append(tok1.id)
								if child in tok1.modifiers:
									tok1.modifiers.remove(child)
									child.modifiers.append(tok1)
								# Check if any other non-link parents need to be re-routed to the new head
								for tok_to_rewire in conll_tokens[tokoffset + 1:]:
									if tok_to_rewire.original_head == tok1.id and tok_to_rewire.head != child.id and tok_to_rewire.id != child.id:
										tok_to_rewire.head = child.id
										# Also add the rewired child func
										if tok_to_rewire.func not in child.child_funcs:
											child.child_funcs.append(tok_to_rewire.func)
										# Rewire modifiers
										if tok_to_rewire not in child.modifiers and lex.filters["mod_func"].match(tok_to_rewire.func) is not None:
											child.modifiers.append(tok_to_rewire)
										if child in tok_to_rewire.modifiers:
											tok_to_rewire.modifiers.remove(child)

								# Only do this for the first subordinate markable head found by traversing right to left
								break

			# Try to construct a longer stop candidate starting with each token in the sentence, max length 5 tokens
			stop_candidate = ""
			for tok2 in conll_tokens[int(tok1.id):min(len(conll_tokens),int(tok1.id)+4)]:
				stop_candidate += tok2.text + " "
				if stop_candidate.strip().lower() in lex.stop_list:  # Stop list matched, flag tokens as impossible markable heads
					for tok3 in conll_tokens[int(tok1.id):int(tok2.id) + 1]:
						stop_ids[tok3.id] = True

		# Find last-first name combinations
		for tok1 in conll_tokens[tokoffset + 1:-1]:
			tok2 = conll_tokens[int(tok1.id) + 1]
			first_name_candidate = tok1.text.title() if tok1.text.isupper() else tok1.text
			last_name_candidate = tok2.text.title() if tok2.text.isupper() else tok2.text
			if not lex.filters["cap_names"] or (first_name_candidate[0].isupper() and last_name_candidate[0].isupper()):
				if first_name_candidate in lex.first_names and last_name_candidate in lex.last_names and tok1.head == tok2.id:
					stop_ids[tok1.id] = True
		# Allow one intervening token, e.g. for middle initial
		for tok1 in conll_tokens[tokoffset + 1:-2]:
			tok2 = conll_tokens[int(tok1.id) + 2]
			first_name_candidate = tok1.text.title() if tok1.text.isupper() else tok1.text
			middle_name_candidate = conll_tokens[int(tok1.id) + 1].text.title() if tok1.text.isupper() else conll_tokens[int(tok1.id) + 1].text
			last_name_candidate = tok2.text.title() if tok2.text.isupper() else tok2.text
			if not lex.filters["cap_names"] or (first_name_candidate[0].isupper() and last_name_candidate[0].isupper()):
				if first_name_candidate in lex.first_names and last_name_candidate in lex.last_names and tok1.head == tok2.id and (re.match(r'^[A-Z]\.$',middle_name_candidate) or middle_name_candidate in lex.first_names):
					stop_ids[tok1.id] = True

		# Expand children list recursively into descendants for this sentence
		for parent_key in children:
			if int(parent_key) > tokoffset and int(parent_key) <= int(conll_tokens[-1].id):
				descendants[parent_key] = get_descendants(parent_key, children, [], self.sent_num, conll_tokens)

		keys_to_pop = []

		# Find markables
		for tok in conll_tokens[tokoffset + 1:]:
			# Markable heads should match specified pos or pos+func combinations,
			# ruling out stop list items with appropriate functions
			if tok.text == lex.debug["ana"]:
				a=5
			# TODO: consider switch for lex.filters["stop_func"].match(tok.func)
			if ((lex.filters["mark_head_pos"].match(tok.pos) is not None and lex.filters["mark_forbidden_func"].match(tok.func) is None) or
					pos_func_combo(tok.pos, tok.func, lex.filters["pos_func_heads"])) and not (stop_ids[tok.id]):
				this_markable = make_markable(tok, conll_tokens, descendants, tokoffset, sentence, keys_to_pop, lex)
				if this_markable is not None:
					mark_candidates_by_head[tok.id] = this_markable

				# Check whether this head is the beginning of a coordination and needs its own sub-markable too
				make_submark = False
				submark_id = ""
				submarks = []
				cardi=0
				for child_id in children[tok.id]:
					child = conll_tokens[int(child_id)]
					if child.coordinate:
						# Coordination found - make a small markable for just this first head without coordinates
						make_submark = True
						# Remove the coordinate children from descendants of small markable head
						if child.id in descendants:
							for sub_descendant in descendants[child.id]:
								if tok.id in descendants:
									if sub_descendant in descendants[tok.id]:
										descendants[tok.id].remove(sub_descendant)
						if tok.id in descendants:
							if child.id in descendants[tok.id]:
								descendants[tok.id].remove(child.id)
						# Build a composite id for the large head from coordinate children IDs separated by underscore
						submark_id += "_" + child.id
						cardi+=1
						submarks.append(child.id)
				if make_submark:
					submarks.append(tok.id)
					# Assign aggregate/coordinate agreement class to large markable if desired
					# Remove coordination tokens, such as 'and', 'or' based on coord_func setting
					for child_id in children[tok.id]:
						child = conll_tokens[int(child_id)]
						if lex.filters["coord_func"].match(child.func):
							if child.id in descendants[tok.id]:
								descendants[tok.id].remove(child.id)

					# Make the small markable and recall the big markable
					mark_candidates_by_head[tok.id].cardinality=cardi+1
					big_markable = mark_candidates_by_head[tok.id]
					small_markable = make_markable(tok, conll_tokens, descendants, tokoffset, sentence, keys_to_pop, lex)
					big_markable.submarks = submarks[:]
					if lex.filters["aggregate_agree"] != "_":
						big_markable.agree = lex.filters["aggregate_agree"]
						big_markable.agree_certainty = "coordinate_aggregate_plural"
						big_markable.coordinate = True

					# Switch the id's so that the big markable has the 1_2_3 style id, and the small has just the head id
					# Check that big_markable's coordinate heads are all eligible
					if all([is_eligible_submark_head(conll_tokens[int(m)]) for m in big_markable.submarks]):
						mark_candidates_by_head[tok.id + submark_id] = big_markable
					mark_candidates_by_head[tok.id] = small_markable
					big_markable = None
					small_markable = None


		# Check for atomicity and remove any subsumed markables if atomic
		for mark_id in mark_candidates_by_head:
			mark = mark_candidates_by_head[mark_id]
			if mark.end > mark.start:  # No atomicity check if single token
				# Check if the markable has a modifier based entity guess
				modifier_based_entity = recognize_entity_by_mod(mark, lex, True)
				# Consider for atomicity if in atoms or has @ modifier, but not if it's a single token or a coordinate markable
				if (is_atomic(mark, lex.atoms, lex) or ("@" in modifier_based_entity and "_" not in mark_id)) and mark.end > mark.start:
					for index in enumerate(mark_candidates_by_head):
						key = index[1]
						# Note that the key may contain underscores if it's a composite, but those can't be atomic
						if key != mark.head.id and mark.start <= int(re.sub('_.*','',key)) <= mark.end and '_' not in key:
							if lex.filters["pronoun_pos"].match(conll_tokens[int(re.sub('_.*','',key))].pos) is None:  # Make sure we're not removing a pronoun
								keys_to_pop.append(key)
				elif len(modifier_based_entity) > 1:
					stoplist_prefix_tokens(mark, lex.entity_mods, keys_to_pop)
			# Check for whole markable only exclusion
			if mark.text + "@" in lex.stop_list:
				keys_to_pop.append(mark_id)

		for key in keys_to_pop:
			mark_candidates_by_head.pop(key, None)

		processed_marks = len(markables)
		for mark_id in mark_candidates_by_head:
			mark = mark_candidates_by_head[mark_id]
			self.analyze_markable(mark, lex)

			self.markcounter += 1
			self.groupcounter += 1
			this_markable = Markable("referent_" + str(self.markcounter), mark.head, mark.form,
									 mark.definiteness, mark.start, mark.end, mark.text, mark.core_text, mark.entity, mark.entity_certainty,
									 mark.subclass, "new", mark.agree, mark.sentence, "none", "none", self.groupcounter,
									 mark.alt_entities, mark.alt_subclasses, mark.alt_agree,mark.cardinality,mark.submarks,mark.coordinate,mark.agree_certainty)
			this_markable.get_dep_freqs(lex)

			markables.append(this_markable)
			markables_by_head[mark_id] = this_markable
			markstart_dict[this_markable.start].append(this_markable)
			markend_dict[this_markable.end].append(this_markable)

		for current_markable in markables[processed_marks:]:
			# DEBUG POINT
			if current_markable.text == lex.debug["ana"]:
				a=5
				if current_markable.text == lex.debug["ante"]:
					a=5
			# Revise coordinate markable entities now that we have resolved all of their constituents
			if len(current_markable.submarks) > 0:
				assign_coordinate_entity(current_markable,markables_by_head)
			if antecedent_prohibited(current_markable, conll_tokens, lex) or (current_markable.definiteness == "indef" and lex.filters["apposition_func"].match(current_markable.head.func) is None and not lex.filters["allow_indef_anaphor"]):
				antecedent = None
			elif (current_markable.definiteness == "indef" and lex.filters["apposition_func"].match(current_markable.head.func) is not None and not lex.filters["allow_indef_anaphor"]):
				antecedent, propagation = find_antecedent(current_markable, markables, lex, "appos")
			else:
				antecedent, propagation = find_antecedent(current_markable, markables, lex)
			if antecedent is not None:
				if int(antecedent.head.id) < int(current_markable.head.id) or 'invert' in propagation:
					# If the rule specifies to invert
					if 'invert' in propagation:
						temp = antecedent
						antecedent = current_markable
						current_markable = temp
					current_markable.antecedent = antecedent
					current_markable.group = antecedent.group

					# Check for apposition function if both markables are in the same sentence
					if lex.filters["apposition_func"].match(current_markable.head.func) is not None and \
							current_markable.sentence.sent_num == antecedent.sentence.sent_num:
						current_markable.coref_type = "appos"
					elif current_markable.form == "pronoun":
						current_markable.coref_type = "ana"
					elif current_markable.coref_type == "none":
						current_markable.coref_type = "coref"
					current_markable.infstat = "giv"
				else:  # Cataphoric match
					current_markable.antecedent = antecedent
					antecedent.group = current_markable.group
					current_markable.coref_type = "cata"
					current_markable.infstat = "new"
			elif current_markable.form == "pronoun":
				current_markable.infstat = "acc"
			else:
				current_markable.infstat = "new"

			if current_markable.agree is not None and current_markable.agree != '':
				lex.last[current_markable.agree] = current_markable
			else:
				pass






# Requirements:
# spacy + English model
# xrenner (pip install xrenner)
# depedit (pip install depedit)

from spacy.en import English

# Part 1: Use spacy to get a dependency parse of the text
text = u"The CEO and the taxi driver sat down. His employees joined them."
parser = English(entity=False,load_vectors=False,vectors_package=False)
parsed = parser(text)


# Part 2: xrenner expects Stanford typed dependencies
# We can get these by running the Stanford Parser with 'basic' dependencies in 'conllx' format
# Or we can use spacy, slightly rewire the results and labels, and construct conll format, as follows:
parsed_string=""
toks = 0
prev_s_toks = 0
for sent in parsed.sents:
	toks = 0
	for index, token in enumerate(sent):
		if token.head.i+1 - prev_s_toks == token.i+1 - prev_s_toks: # Spacy represents root as self-dominance, revert to 0
			head_id = 0
		else:
			head_id = token.head.i + 1 - prev_s_toks

		line = [str(token.i+1 - prev_s_toks),token.orth_,token.lemma_,token.tag_,token.pos_,"_", str(head_id), token.dep_, "_","_"]
		parsed_string += "\t".join(line) + "\n"
		toks += 1
	prev_s_toks += toks
	parsed_string += "\n"

# The parse string now looks like this:
print "Original spacy parse, in conll format:"
print parsed_string

# Part 3: editing the labels to be Stanford-like
# This part uses depedit: a module to do 'find and replace' in dependency trees
# To learn more about how it works, see the user guide at: http://corpling.uis.georgetown.edu/depedit

from depedit import DepEdit

config =  "func=/ROOT/\tnone\t#1:func=root\n"
config += "func=/relcl/\tnone\t#1:func=rcmod\n"
config += "func=/nummod/\tnone\t#1:func=num\n"
config += "lemma=/be/&func=/(.*)/;func=/nsubj/;func=/attr/;text=/.*/\t#1>#2;#1>#3;#4>#1\t#4>#3;#3>#1;#3>#2;#1:func=cop;#3:func=$1\n"
config += "lemma=/be/&func=/root/;func=/nsubj/;func=/attr/\t#1>#2;#1>#3\t#3>#1;#3>#2;#1:func=cop;#3:func=root\n"

deped = DepEdit(config.split("\n"))

edited = deped.run_depedit(parsed_string.split("\n"))

# The parse string now looks like this:
print "Edited parse, after depedit:"
print edited


# Part 4: actual coref stuff
from xrenner.modules.xrenner_xrenner import Xrenner

# Load a model to initialize the Xrenner object with language specific lexical information
xrenner = Xrenner(model="C:\\eng.xrm",override="GUM")  # If you omit the gum override, you will get the OntoNotes schema, which ignores singleton mentions among other things

# Output the analysis in conll coref format - you can also try "html" or "sgml", among other options
print xrenner.analyze(edited, "conll")




####### Object Model Examples ########
# Heres how to access some properties of the object model once the analysis has been run:

all_markables = xrenner.markables
fourth_markable = all_markables[3]  # This is "His"
print fourth_markable  # Get a textual summary of this markable

# Get the lemma of the antecedent for fourth markable if it's given
if fourth_markable.infstat == "giv":
	print fourth_markable.antecedent.lemma  # Note that the antecedent is a markable object, and has all the same properties
	# Compare grammatical functions for anaphor and antecedent if they are not in the same sentence
	if fourth_markable.sentence.sent_num == fourth_markable.antecedent.sentence.sent_num:
		print "Antecedent is in same sentence"
	else:
		if fourth_markable.func == fourth_markable.antecedent.func:
			print "The anaphor retains the same function: " + fourth_markable.func
		else:
			print "Anaphor function is: " + fourth_markable.func + " but antecedent is: " + fourth_markable.antecedent.func
else:
	print "The markable's information status is " + fourth_markable.infstat

# See the slides for some other markable properties you can play with - there are many more



## PS - xrenner can also be fed a text file containing a parse from the file system like this:
#print xrenner.analyze("C:\\parses\\example_parse.conll10","conll")


"""
Basic classes for parsed tokens, markables and sentences.

Author: Amir Zeldes
"""

import sys, re
from math import log
from collections import OrderedDict, defaultdict

class ParsedToken:
	def __init__(self, tok_id, text, lemma, pos, morph, head, func, sentence, modifiers, child_funcs, child_strings, lex, quoted=False, head2="_", func2="_"):
		self.id = tok_id
		self.text = text.strip()
		self.text_lower = text.lower()
		self.pos = pos
		if lemma != "_" and lemma != "--":
			self.lemma = lemma.strip()
		else:
			self.lemma = lex.lemmatize(self)
		self.morph = morph
		if morph != "_" and morph != "--" and morph != "":
			self.morph = lex.process_morph(self)

		self.head = head
		self.original_head = head
		self.func = func
		self.head2 = head2
		self.func2 = func2
		self.sentence = sentence
		self.modifiers = modifiers
		self.child_funcs = child_funcs
		self.child_strings = child_strings
		self.quoted = quoted
		self.coordinate = False
		self.head_text = ""
		self.lex = lex  # Pointer to lex object
		self.lemma_freq = 0.0
		self.negated = False
		self.neg_parent = False

	def __repr__(self):
		return str(self.text) + " (" + str(self.pos) + "/" + str(self.lemma) + ") " + "<-" + str(self.func) + "- " + str(self.head_text)

class Markable:

	# Properties refering to markable head, not markable itself
	tok_props = {"negated", "neg_parent", "pos", "lemma", "morph", "func", "quoted", "modifiers", "child_funcs", "child_strings",
	 "agree", "doc_position", "sent_position", "head_text", "head_pos", "lemma_freq"}

	def __init__(self, mark_id, head, form, definiteness, start, end, text, core_text, entity, entity_certainty, subclass, infstat, agree, sentence,
				 antecedent, coref_type, group, alt_entities, alt_subclasses, alt_agree, cardinality=0, submarks=[], coordinate=False, agree_certainty=""):
		self.id = mark_id
		self.head = head
		self.form = form
		self.definiteness = definiteness
		self.start = start
		self.end = end
		self.text = text.strip()
		self.core_text = core_text.strip()  # Holds markable text before any extensions or manipulations
		self.first = self.core_text.split(" ")[0]
		self.last = self.core_text.split(" ")[-1]
		self.entity = entity
		self.subclass = subclass
		self.infstat = infstat
		self.agree = agree
		self.agree_certainty = agree_certainty
		self.sentence = sentence
		self.antecedent = antecedent
		self.coref_type = coref_type
		self.group = group
		self.non_antecdent_groups = set()
		self.entity_certainty = entity_certainty
		self.isa_partner_head = ""  # Property to hold isa match; once saturated, no other lexeme may form isa link

		# Alternate agreement, subclass and entity lists:
		self.alt_agree = alt_agree
		self.alt_entities = alt_entities
		self.alt_subclasses = alt_subclasses

		self.cardinality=cardinality
		self.submarks = submarks
		self.coordinate = coordinate

		self.length = self.text.strip().count(" ") + 1
		self.mod_count = len(self.head.modifiers)

		# Dictionaries to fill when getting dependency and similarity based frequencies
		self.entity_dep_scores = defaultdict(int)
		self.entity_sim_dep_scores = defaultdict(int)
		self.lex_dep_scores = defaultdict(int)
		self.lex_sim_dep_scores = defaultdict(int)

	def has_child_func(self, func):
		if "*" in func: # func substring, do not delimit function
			return func in self.child_func_string
		else: # Exact match, delimit with ";"
			return ";" + func + ";" in self.child_func_string

	def get_dep_freqs(self,lex):

		#DEBUG POINT
		if self.text == lex.debug["ana"]:
			pass

		use_entity_deps = True
		use_lex_deps = True
		if "ablations" in lex.debug:
			if "no_entity_dep" in lex.debug["ablations"]:
				use_entity_deps = False
		use_hasa = True
		if "ablations" in lex.debug:
			if "no_hasa" in lex.debug["ablations"]:
				use_hasa = False

		anaphor_parent = self.head.head_text
		if anaphor_parent in lex.entity_deps and use_entity_deps:
			if self.head.func in lex.entity_deps[anaphor_parent]:
				for entity in lex.entity_deps[anaphor_parent][self.head.func]:
					self.entity_dep_scores[entity] = lex.entity_deps[anaphor_parent][self.head.func][entity]
		found = False
		if anaphor_parent in lex.similar and use_entity_deps:
			for sim in lex.similar[anaphor_parent]:
				if sim in lex.entity_deps and not found:
					if self.head.func in lex.entity_deps[sim]:
						found = True
						for entity in lex.entity_deps[sim][self.head.func]:
							self.entity_sim_dep_scores[entity] = lex.entity_deps[sim][self.head.func][entity]
		if anaphor_parent in lex.lex_deps and use_lex_deps:
			if self.head.func in lex.lex_deps[anaphor_parent]:
				for lexeme in lex.lex_deps[anaphor_parent][self.head.func]:
					self.lex_dep_scores[lexeme] = lex.lex_deps[anaphor_parent][self.head.func][lexeme]
		if anaphor_parent in lex.similar and use_entity_deps:
			for sim in lex.similar[anaphor_parent]:
				if sim in lex.lex_deps:
					if self.head.func in lex.lex_deps[sim]:
						for lexeme in lex.lex_deps[sim][self.head.func]:
							self.entity_sim_dep_scores[lexeme] = lex.lex_deps[sim][self.head.func][lexeme]

	def __repr__(self):
		agree = "no-agr" if self.agree == "" else self.agree
		defin = "no-def" if self.definiteness == "" else self.definiteness
		card = "no-card" if self.cardinality == 0 else self.cardinality
		func = "no-func" if self.head.func == "" else self.head.func
		return str(self.entity) + "/" + str(self.subclass) + ': "' + self.text + '" (' + agree + "/" + defin + "/" + func + "/" + str(card) + ")"


	def extract_features(self, lex, antecedent=None, candidate_list=[], dump_position=False):
		"""
		Function to generate feature representation of markables or markable-antecedent pairs for classifiers

		:param lex: the LexData object with gazetteer information and model settings
		:param antecedent: The antecedent Markable potentially coreferring to self
		:param candidate_list: The list of candidate markables under consideration, used to extract cohort size
		:param dump_position: Whether document name + token positions are dumped for each markable to compare to gold
		:return: dictionary of markable properties
		"""

		out_dict = OrderedDict()
		if dump_position:
			out_dict["position"] = str(self.start)+"-"+str(self.end)+";"+str(antecedent.start)+"-"+str(antecedent.end)
		out_dict["docname"] = lex.docname

		# TODO: Make genre representation configurable
		# By convention, genre is taken from first 4 chars of file, but for corpora like GUM use part between underscores
		if lex.docname.startswith("GUM_") or lex.docname.lower().startswith("autogum_") or lex.docname.lower().startswith("amalgum_"):
			out_dict["genre"] = lex.docname.split("_")[1]
		elif len(lex.docname) > 4:
			out_dict["genre"] = lex.docname[:4]
		else:
			out_dict["genre"] = "_"  # Too short to detect genre

		log_props = set([])  # Properties to log-transform before dump, if desired
		bool_props = {"coordinate","quoted","negated","neg_parent"}  # Boolean props to dump as 1 or 0
		thresh_props = {"lemma","head_text"}
		f_threshold = 0  # Minimum frequency for lexical categories in freqs.tab (0 to ignore, else items with lower freq are replace by POS)

		# TODO: May need to lower() head_text, since cap/uncapped version may be in lex/entity dep score lexicon
		anaphor_parent = self.head.head_text
		ana_props = ["lemma","func","head_text","form","pos","agree","start","end","lemma_freq",
						  "cardinality", "definiteness","entity","subclass", "infstat", "coordinate",
						  "length", "mod_count", "doc_position","sent_position","quoted","negated","neg_parent","s_type"]
		ante_props = ["lemma","func","head_text","form","pos","agree","start","end","lemma_freq",
						  "cardinality", "definiteness","entity","subclass", "infstat", "coordinate",
						  "length", "mod_count", "doc_position","sent_position","quoted","negated","neg_parent","s_type"]
		for prop in ana_props:
			val = getattr(self,prop)
			if prop in log_props:
				val = log(val+1)
			elif prop in bool_props:
				val = int(val)
			elif prop in ["lemma"]:
				if lex.freqs[val] < f_threshold:
					val = self.pos
			elif prop in ["head_text"]:
				if lex.freqs[val] < f_threshold:
					val = self.head_pos
			out_dict["n_" + prop] = val if val != "" else "_"

		if antecedent is not None:
			for prop in ante_props:
				val = getattr(antecedent,prop)
				if prop in log_props:
					val = log(val + 1)
				elif prop in bool_props:
					val = int(val)
				elif prop in thresh_props:
					if lex.freqs[val] < f_threshold:
						if prop == "lemma":
							val = antecedent.pos
						elif prop == "head_text":
							val = antecedent.head_pos
				out_dict["t_" + prop] = val if val != "" else "_"

			anaphor = self

			out_dict["d_sent"] = anaphor.sent_num - antecedent.sent_num
			out_dict["d_tok"] = anaphor.start - antecedent.end
			out_dict["d_agr"] = int(anaphor.agree == antecedent.agree)
			out_dict["d_intervene"] = abs(int(re.sub('.*_', '', anaphor.id)) - int(re.sub('.*_', '', antecedent.id)))  # Number of markables between ana and ante
			out_dict["d_cohort"] = len(candidate_list)
			out_dict["d_modcount"] = anaphor.mod_count - antecedent.mod_count
			out_dict["d_samemods"] = len(list(set([m.lemma for m in anaphor.head.modifiers]) & set([m.lemma for m in antecedent.head.modifiers])))

			hasa = 0
			use_hasa = True
			if "ablations" in lex.debug:
				if "no_hasa" in lex.debug["ablations"]:
					use_hasa = False
			if use_hasa:
				if antecedent.head.text in lex.hasa and lex.filters["possessive_func"].search(self.func) is not None:  # Text based hasa
					if anaphor_parent in lex.hasa[antecedent.head.text]:
						hasa = lex.hasa[antecedent.head.text][anaphor_parent]
				elif antecedent.head.lemma in lex.hasa and lex.filters["possessive_func"].search(self.func) is not None:  # Lemma based hasa
					if anaphor_parent in lex.hasa[antecedent.head.lemma]:
						hasa = lex.hasa[antecedent.head.lemma][anaphor_parent]

			out_dict["d_hasa"] = hasa
			out_dict["d_entidep"] = self.entity_dep_scores[antecedent.entity]
			out_dict["d_entisimdep"] = self.entity_sim_dep_scores[antecedent.entity]
			out_dict["d_lexdep"] = self.lex_dep_scores[antecedent.head.text]
			out_dict["d_lexsimdep"] = self.lex_sim_dep_scores[antecedent.head.text]
			out_dict["d_sametext"] = int(anaphor.text == antecedent.text)
			out_dict["d_samelemma"] = int(anaphor.lemma == antecedent.lemma)

			out_dict["d_doclen"] = int(anaphor.head.lex.token_count)


			# Check if one markable head is the dependency parent of the other
			if antecedent.head.head == anaphor.head.id:
				out_dict["d_parent"] = 1
			elif anaphor.head.head == antecedent.head.id:
				out_dict["d_parent"] = -1
			else:
				out_dict["d_parent"] = 0

			if self.speaker == "" and antecedent.speaker == "":
				out_dict["d_speaker"] = 0
			elif self.speaker == antecedent.speaker:
				out_dict["d_speaker"] = 1
			else:
				out_dict["d_speaker"] = -1

		return out_dict

	def __getattr__(self, item):
		# Convenience methods to access head token and containing sentence
		if item in self.tok_props:
			return getattr(self.head,item)
		elif item == "text_lower":
			if self.coordinate:  # If this is a coordinate markable return lower case core_text
				return self.core_text.lower()
			else:  # Otherwise return lower text of head token
				return getattr(self.head, item)
		elif item in ["mood", "speaker", "sent_num", "s_type"]:
			return getattr(self.sentence,item)
		elif item == "child_func_string":
			# Check for cached child_func_string
			if "child_func_string" not in self.__dict__:  # Convenience property to store semi-colon separated child funcs of head token
				# Assemble if not yet cached
				if len(self.head.child_funcs) > 1:
					self.child_func_string = ";" + ";".join(self.head.child_funcs) + ";"
				else:
					self.child_func_string = "_"
			return self.child_func_string
		else:
			raise AttributeError("No attribute: " + str(item))


class Sentence:
	def __init__(self, sent_num, start_offset, mood="", speaker=""):
		self.sent_num = sent_num
		self.start_offset = start_offset
		self.mood = mood
		self.speaker = speaker
		self.token_count = 0
		self.s_type = "_"  # Initial type, will be overwritten if s_type annotation is found in input

	def __repr__(self):
		mood = "(no mood info)" if self.mood == "" else self.mood
		speaker = "(no speaker info)" if self.speaker == "" else self.speaker
		return "S" + str(self.sent_num) + " from T" + str(self.start_offset + 1) + ", mood: " + mood  + ", speaker: " + speaker + ", type: " + self.s_type


def get_descendants(parent, children_dict, seen_tokens, sent_num, conll_tokens):
	my_descendants = []
	my_descendants += children_dict[parent]
	for child in children_dict[parent]:
		if child in seen_tokens:
			if sys.version_info[0] < 3:
				sys.stderr.write("\nCycle detected in syntax tree in " + conll_tokens[int(parent)].lex.docname + " in sentence " + str(sent_num) + " (child of token: '" + conll_tokens[int(parent)].text.encode("utf8") + "')\n")
			else:
				sys.stderr.write("\nCycle detected in syntax tree in " + conll_tokens[int(parent)].lex.docname + " in sentence " + str(sent_num) + " (child of token: '" + conll_tokens[int(parent)].text + "')\n")
			sys.exit("Exiting due to invalid input\n")
		else:
			seen_tokens += [child]
	for child in children_dict[parent]:
		if child in children_dict:
			my_descendants += get_descendants(child, children_dict, seen_tokens, sent_num, conll_tokens)
	return my_descendants



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
xrenner - eXternally configurable REference and Non-Named Entity Recognizer
xrenner.py
Main controller script for entity recognition and coreference resolution
Author: Amir Zeldes
"""

import argparse, sys, io, os
from modules.xrenner_xrenner import Xrenner
from glob import glob
from multiprocessing import Process, Value, Lock, current_process, Array
import ctypes
from math import ceil

__version__ = "2.1.0"
xrenner_version = "xrenner V" + __version__

sys.dont_write_bytecode = True

if sys.version_info[0] > 2:
	PY3 = True
else:
	PY3 = False
	reload(sys)
	sys.setdefaultencoding('utf8')


class Counter(object):
	"""
	Counter to synchronize progress in multiple processes. Set lock to False
	for single process, especially for running from a subprocess.
	Note that for verbose behavior, lock must be True, and subprocess invocation
	may fail.
	"""
	def __init__(self, initval=0, lock=True):
		if lock:
			self.docs = Value('i', initval)
			self.sents = Value('i', initval)
			self.toks = Value('i', initval)
			if PY3:
				self.dump_headers = Array(ctypes.c_char, b"_" * 3000)  # Value(ctypes.c_char_p, "")
			else:
				self.dump_headers = Array(ctypes.c_char,"_"*3000)# Value(ctypes.c_char_p, "")
			self.lock = Lock()
		else:
			self.docs = initval
			self.sents = initval
			self.toks = initval
			self.lock = False
			self.dump_headers = ""

	def increment(self,docs,sents,toks):
		if not self.lock:
			self.docs += docs
			self.sents += sents
			self.toks += toks
		else:
			with self.lock:
				self.docs.value += docs
				self.sents.value += sents
				self.toks.value += toks

	def set_string(self,new_val):
		if not self.lock:
			self.dump_headers = new_val
		else:
			with self.lock:
				self.dump_headers.value = new_val

	def get_string(self):
		if not self.lock:
			return self.dump_headers
		else:
			with self.lock:
				return self.dump_headers.value

	def value(self):
		if not self.lock:
			return (self.docs, self.sents, self.toks)
		else:
			with self.lock:
				return (self.docs.value,self.sents.value,self.toks.value)


def rreplace(s, old, new, occurrence):
	li = s.rsplit(old, occurrence)
	return new.join(li)

def xrenner_worker(data,options,total_docs,counter):
	tokens = 0
	sentences = 0

	model = options.model
	override = options.override
	xrenner = Xrenner(model, override, options.rulebased, options.noseq)

	if options.dump is not None:
		xrenner.lex.procid = str(current_process().ident)
		xrenner.lex.dump = io.open(rreplace(options.dump,".", xrenner.lex.procid+".",1),'w',encoding="utf8",newline="\n")
	else:
		xrenner.lex.procid = ""
		xrenner.lex.dump = None

	if options.oracle is not None:
		xrenner.lex.read_oracle(options.oracle)

	for file_ in data:

		xrenner.lex.dump_types = set([])  # Empty set of dump rows to avoid duplicates
		output = xrenner.analyze(file_, options.format)
		tokens += len(xrenner.conll_tokens)-1
		sentences += xrenner.sent_num-1

		if options.format == "none":
			pass
		elif options.format != "paula":
			if len(data) > 1:
				if options.format == "webanno":
					extension = "xmi"
				elif options.format == "webannotsv":
					extension = "tsv"
				else:
					extension = options.format
				outfile = xrenner.docname + "." + extension
				handle = io.open(outfile, 'w', encoding="utf8")
				handle.write(output)
				handle.close()
			else:
				if PY3:
					sys.stdout.buffer.write(output.encode("utf8"))
				else:
					print(output.encode("utf8"))

		counter.increment(1,xrenner.sent_num-1,len(xrenner.conll_tokens)-1)
		docs, sents, toks = counter.value()
		if options.verbose and options.oracle is not None:
			sys.stderr.write("Used oracle for " + str(xrenner.lex.oracle_counters[2]) + " entities, of which " + str(xrenner.lex.oracle_counters[0]) + " had spans in oracle and " + str(xrenner.lex.oracle_counters[1]) + " were different types than xrenner pred\n")

		if options.verbose and len(data) > 1:
			sys.stderr.write("Document " + str(docs) + "/" + str(total_docs) + ": " +
								 "Processed " + str(len(xrenner.conll_tokens)-1) + " tokens in " + str(xrenner.sent_num-1) + " sentences.\n")

	if options.dump is not None:
		xrenner.lex.dump.close()
		if PY3:
			counter.set_string(bytes("\t".join(xrenner.lex.dump_headers),encoding="utf8"))
		else:
			counter.set_string("\t".join(xrenner.lex.dump_headers))


if __name__ == "__main__":

	parser = argparse.ArgumentParser()
	parser.add_argument('-o', '--output', action="store", dest="format", default="sgml", help="output format, default: sgml",
						choices=["sgml", "html", "paula", "webanno", "webannotsv", "conll", "onto", "unittest", "none"])
	parser.add_argument('-m', '--model', action="store", dest="model", default="eng", help="input model directory name, in models/")
	parser.add_argument('-x', '--override', action="store", dest="override", default=None, help="specify a section in the model's override.ini file with alternative settings")
	parser.add_argument('-r', '--rulebased', action="store_true", help="run model without machine learning classifiers")
	parser.add_argument('-v', '--verbose', action="store_true", help="output run time and summary")
	parser.add_argument('-t', '--test', action="store_true", dest="test", help="run unit tests and quit")
	parser.add_argument('-p', '--procs', type=int, choices=range(1,17), dest="procs", help="number of processes for multithreading", default=2)
	parser.add_argument('-d', '--dump', action="store", dest="dump", help="file to dump individual analyses into", default=None)
	parser.add_argument('file', action="store", help="input file name to process")
	parser.add_argument('--oracle', action='store', help="file with oracle entity predictions")
	parser.add_argument('--noseq', action='store_true', help="do not use sequence tagger for entity classification")
	parser.add_argument('--version', action='version', version=xrenner_version, help="show xrenner version number and quit")

	total_docs = 0

	# Check if -t is invoked and run unit tests instead of parsing command line
	if len(sys.argv) > 1 and sys.argv[1] in ["-t", "--test"]:
		import unittest
		import modules.xrenner_test
		suite = unittest.TestLoader().loadTestsFromModule(modules.xrenner_test)
		unittest.TextTestRunner().run(suite)
	# Not a test run, parse command line as usual
	else:
		options = parser.parse_args()
		procs = options.procs
		if options.verbose:
			import modules.timing
			sys.stderr.write("\nReading language model...\n")

		data = glob(options.file)
		if data == []:
			sys.stderr.write("\nCan't find input at " + options.file +"\nAborting\n")
			sys.exit()
		if not isinstance(data, list):
			split_data = [data]
		else:
			if len(data) < procs:  # Do not use more processes than files to process
				procs = len(data)
			chunk_size = int(ceil(len(data)/float(procs)))
			split_data = [data[i:i + chunk_size] for i in range(0, len(data), chunk_size)]

		if procs > 1 or options.verbose:
			lock = True
		else:
			lock = True
		counter = Counter(0,lock=lock)
		jobs = []
		dump_files = []

		if options.dump is not None:
			if "." not in options.dump:
				options.dump += ".tab"

		for sublist in split_data:
			p = Process(target=xrenner_worker, args=(sublist,options,len(data), counter))
			jobs.append(p)
			p.start()
			if options.dump is not None:
				dump_files.append(rreplace(options.dump, ".", str(p.ident)+".",1))
		for j in jobs:
			j.join()

		total_docs, total_sentences, total_tokens = counter.value()

		if options.verbose:
			sys.stderr.write("="*40 + "\n")
			sys.stderr.write("Processed " + str(total_tokens) + " tokens in " + str(total_sentences) + " sentences.\n")

		if len(dump_files) > 0:
			# Merge dump files from multiple processes
			sys.stderr.write("Collating dump data ... \n")
			with io.open(options.dump,'w',encoding="utf8",newline="\n") as wfd:
				if lock:
					if PY3:
						headers = counter.dump_headers.value.decode("utf8")
					else:
						headers = counter.dump_headers.value
				else:
					headers = counter.dump_headers
				wfd.write(headers + "\n")
			with io.open(options.dump,'a',encoding="utf8") as wfd:
				for f in dump_files:
					with io.open(f,'r', encoding="utf8") as fd:
						temp = fd.read()
						wfd.write(temp)
			for f in dump_files:
				os.remove(f)
			sys.stderr.write("Dump written to "+options.dump+"\n")




"""
modules/xrenner_sequence.py

Adapter class to accommodate sequence labeler
  * Supplies a uniform predict_proba() method
  * Reads serialized models
  * Compatible with flair embeddings

Author: Amir Zeldes
"""

import sys, os

script_dir = os.path.dirname(os.path.realpath(__file__)) + os.sep

PY3 = sys.version_info[0] > 2

class StdOutFilter(object):
	def __init__(self): #, strings_to_filter, stream):
		self.stream = sys.stdout
		self.stdout = sys.stdout

	def __getattr__(self, attr_name):
		return getattr(self.stream, attr_name)

	def write(self, data):
		output = []
		lines = data.split("\n")
		for line in lines:
			if "Epoch: " in line or (" f:" in line and "Test" not in line):
				output.append(line)
		if len(output)>0:
			data = "\n".join(output) + "\n"
			self.stream.write("RNN log - " + data.strip() + "\n")
			self.stream.flush()

	def flush(self):
		pass
		#self.stream.flush()

	def start(self):
		sys.stdout = self

	def end(self):
		sys.stdout = self.stdout

p = StdOutFilter()
p.start()

# Silently import flair/torch
from flair.data import Sentence
from flair.models import SequenceTagger
import torch

p.end()


class Sequencer:
	def __init__(self, model_path=None):
		if model_path is None:
			model_path = script_dir + ".." + os.sep + "models" + os.sep + "_sequence_taggers" + os.sep + "eng_flair_nner_distilbert.pt"
		elif os.sep not in model_path:  # Assume this is a file in models/_sequence_taggers
			model_path = script_dir + ".." + os.sep + "models" + os.sep + "_sequence_taggers" + os.sep + model_path
		if not os.path.exists(model_path):
			sys.stderr.write("! Sequence tagger model file missing at " + model_path + "\n")
			sys.stderr.write("! Add the model file or use get_models.py to obtain built-in models\nAborting...\n")
			quit()
		self.tagger = SequenceTagger.load(model_path)

	def clear_embeddings(self, sentences, also_clear_word_embeddings=False):
		"""
		Clears the embeddings from all given sentences.
		:param sentences: list of sentences
		"""
		for sentence in sentences:
			sentence.clear_embeddings(also_clear_word_embeddings=also_clear_word_embeddings)

	def predict_proba(self, sentences):
		"""
		Predicts a list of class and class probability tuples for every token in a list of sentences
		:param sentences: list of space tokenized sentence strings
		:return: the list of sentences containing the labels
		"""

		# Sort sentences and keep order
		sents = [(len(s.split()),i,s) for i, s in enumerate(sentences)]
		sents.sort(key=lambda x:x[0], reverse=True)
		sentences = [s[2] for s in sents]

		preds = self.tagger.predict(sentences)

		# sort back
		sents = [tuple(list(sents[i]) + [s]) for i, s in enumerate(preds)]
		sents.sort(key=lambda x:x[1])
		sents = [s[3] for s in sents]

		output = []
		for s in sents:
			for tok in s.tokens:
				output.append((tok.tags['ner'].value.replace("S-","").replace("B-","").replace("I-","").replace("E-",""),
							   tok.tags['ner'].score))

		return output


if __name__ == "__main__":
	c = Sequencer()
	x = c.predict_proba(["Mary had a little lamb","Her fleece was white as snow .","I joined Intel in the Age of Knives"])
	print(x)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
# Python port of TT2CoNLL.pl
#
# this script takes a corpus in CWB SGML format
# and converts it into the CoNLL input format using
# the POS tag column (column 2) as both the fine
# and coarse grained tag. SGML elements are deleted.
#
# Computer	NN	Computer
#
# becomes
#
# 1	Computer	Computer	NN	NN	_	_	_
#
# The script splits sentences based on a predetermined sentence
# splitting tag, at the moment defaulting to $. (STTS sentence
# punctuation)
#
# usage:
# python tt2conll.py [OPTIONS] <CORPUS>
#
# Options and arguments:
#
# -t <tagname> Use the tag <tagname> to split the corpus into sentences, default is $.
# -x <element> Use the XML element <element> to split the corpus into sentences (e.g. s)
# -h           Print this message and quit
#
# <CORPUS>     A corpus coded in CWB SGML
#
# examples:
# python tt2conll.py -t $. my_corpus.sgml > my_corpus.conll
# python tt2conll.py -t SENT my_corpus.sgml > my_corpus.conll
# python tt2conll.py -x s my_corpus.sgml > my_corpus.conll
# etc.
"""

from argparse import ArgumentParser
import io, sys

PY3 = sys.version_info[0] == 3


def conllize(in_text,tag=None,element=None,no_zero=False, super_mapping=None, ten_cols=False):

	if not PY3:
		if not isinstance(in_text,unicode):
			in_text = unicode(in_text.decode("utf8"))

	xml = False
	if element is not None:
		xml = True

	outlines = []
	counter = 1
	for line in in_text.replace("\r","").split("\n"):
		if len(line.strip()) > 0:
			tabs = line.count("\t")
			lemma = "_"
			pos = "_"
			morph = "_"
			misc = "_"
			if tabs == 0:
				tok = line
			else:
				fields = line.split("\t")
				tok = fields[0]
				pos = fields[1]
				if tabs > 1:
					lemma = fields[2]
				if tabs > 2:
					morph = fields[3]
				if tabs > 3:
					misc = fields[4]
			if not (line.startswith("<") and line.endswith(">")):  # Do not make tokens out of XML elements
				if not ten_cols:
					if no_zero:
						outlines.append("\t".join([str(counter), tok, lemma, pos, pos, morph, "_", "_"]))
					else:
						outlines.append("\t".join([str(counter),tok,lemma,pos,pos,morph,"0","_"]))
				else:
					if no_zero:
						outlines.append("\t".join([str(counter), tok, lemma, pos, pos, morph, "_", "_", "_", misc]))
					else:
						outlines.append("\t".join([str(counter), tok, lemma, pos, pos, morph, "0", "_", "_", misc]))
				counter += 1
			if xml:
				if "</" + element + ">" in line:
					counter = 1
					outlines.append("")
			else:
				if pos == tag:
					counter = 1
					outlines.append("")
	if super_mapping is None:
		return "\n".join(outlines)
	else:
		return add_supertokens(outlines, super_mapping)


def add_supertokens(conll_lines,mapping):

	output = []
	counter = 0

	for line in conll_lines:
		if "\t" in line and not line.startswith("#"):
			tok_id = line.split("\t")[0]
			if "-" in tok_id:
				output.append(line)
				continue
			if counter in mapping:
				super_tok, super_len = mapping[counter]
				super_id = tok_id + "-" + str(int(tok_id)+super_len-1)
				super_line = "\t".join([super_id,super_tok] + ["_"]*8)
				output.append(super_line)
			counter+=1

		output.append(line)

	return "\n".join(output)


if __name__ == "__main__":
	parser = ArgumentParser()
	parser.add_argument("-t","--tagname",action="store",default="$.",help="Use the POS tag <tagname> to split the corpus into sentences, default is $.")
	parser.add_argument("-x","--xml_element",action="store",default=None,help="Use the XML element <element> to split the corpus into sentences (e.g. s)")
	parser.add_argument("infile",action="store",help="file to process")

	opts = parser.parse_args()

	in_data = io.open(opts.infile,encoding="utf8").read().replace("\r","")
	conllized = conllize(in_data,opts.tagname,opts.xml_element)
	print(conllized)




"""
modules/xrenner_propagate.py

Feature propagation module. Propagates entity and agreement features for coreferring markables.

Author: Amir Zeldes
"""

def propagate_entity(markable, candidate, direction="propagate"):
	"""
	Propagate class and agreement features between coreferent markables
	
	:param markable: a Markable object
	:param candidate: a coreferent antecedent Markable object
	:param direction: propagation direction; by default, data can be propagated in either direction from the more certain markable to the less certain one, but direction can be forced, e.g. 'propagate_forward'
	:return: void
	"""
	# Check for rule explicit instructions
	if direction == "propagate_forward":
		markable.entity = candidate.entity
		markable.subclass = candidate.subclass
		markable.entity_certainty = "propagated"
		propagate_agree(candidate, markable)
	elif direction == "propagate_back":
		candidate.entity = markable.entity
		candidate.subclass = markable.subclass
		candidate.entity_certainty = "propagated"
		propagate_agree(markable, candidate)
	else:
		# Prefer nominal propagates to pronoun
		if markable.form == "pronoun" and candidate.entity_certainty != "uncertain":
			markable.entity = candidate.entity
			markable.subclass = candidate.subclass
			propagate_agree(candidate, markable)
			markable.entity_certainty = "propagated"
		elif candidate.form == "pronoun" and markable.entity_certainty != "uncertain":
			candidate.entity = markable.entity
			candidate.subclass = markable.subclass
			candidate.entity_certainty = "propagated"
			propagate_agree(markable, candidate)
		else:
			# Prefer certain propagates to uncertain
			if candidate.entity_certainty == "uncertain":
				candidate.entity = markable.entity
				candidate.subclass = markable.subclass
				candidate.entity_certainty = "propagated"
				propagate_agree(markable, candidate)
			elif markable.entity_certainty == "uncertain":
				markable.entity = candidate.entity
				markable.subclass = candidate.subclass
				markable.entity_certainty = "propagated"
				propagate_agree(candidate, markable)
			else:
				# Prefer to propagate to satisfy alt_entity
				if markable.entity != candidate.entity and markable.entity in candidate.alt_entities:
					candidate.entity = markable.entity
					candidate.subclass = markable.subclass
					candidate.entity_certainty = "certain"
					propagate_agree(markable, candidate)
				elif markable.entity != candidate.entity and candidate.entity in markable.alt_entities:
					markable.entity = candidate.entity
					markable.subclass = candidate.subclass
					markable.entity_certainty = "certain"
					propagate_agree(candidate, markable)
				else:
					# Prefer to propagate backwards
					candidate.entity = markable.entity
					candidate.subclass = markable.subclass
					candidate.entity_certainty = "propagated"
					propagate_agree(markable, candidate)


def propagate_agree(markable, candidate):
	"""
	Progpagate agreement between to markables if one has unknown agreement
	
	:param markable: Markable object
	:param candidate: Coreferent antecdedent Markable object
	:return: void
	"""
	if (candidate.agree == '' or candidate.agree is None) and not (markable.agree == '' or markable.agree is None):
		candidate.agree = markable.agree
	else:
		markable.agree = candidate.agree



import atexit
import sys
from time import clock
from functools import reduce

def secondsToStr(t):
    return "%d:%02d:%02d.%03d" % \
        reduce(lambda ll,b : divmod(ll[0],b) + ll[1:],
            [(t*1000,),1000,60,60])

line = "="*40
def log(s, elapsed=None):
    #print line
    #print secondsToStr(clock()), '-', s
    if elapsed:
        sys.stderr.write("Elapsed time: " + str(elapsed) + "\n")
    sys.stderr.write(line +"\n")

def endlog():
    end = clock()
    elapsed = end-start
    log("End Program", secondsToStr(elapsed))

def now():
    return secondsToStr(clock())

start = clock()

atexit.register(endlog)
#log("Start Program")


#!/usr/bin/python
# -*- coding: utf-8 -*-

import re, sys, io, os, platform
import tempfile
import subprocess
from glob import glob

from rftokenizer import RFTokenizer
try:  # Module usage
	from .lib.xrenner import Xrenner
	from .lib._version import __version__
	from .lib.tt2conll import conllize
	from .lib.append_column import inject_col
	from .lib.sent_split import toks_to_sents
	from .lib.whitespace_tokenize import tokenize as whitespace_tokenize
except ImportError:  # direct script usage
	from lib.xrenner import Xrenner
	from lib._version import __version__
	from lib.tt2conll import conllize
	from lib.append_column import inject_col
	from lib.sent_split import toks_to_sents
	from lib.whitespace_tokenize import tokenize as whitespace_tokenize


PY3 = sys.version_info[0] > 2

if PY3:
	from urllib.request import urlretrieve
else:
	from urllib import urlretrieve

inp = input if PY3 else raw_input

script_dir = os.path.dirname(os.path.realpath(__file__))
lib_dir = script_dir + os.sep + "lib" + os.sep
bin_dir = script_dir + os.sep + "bin" + os.sep
data_dir = script_dir + os.sep + "data" + os.sep
model_dir = script_dir + os.sep + "models" + os.sep
parser_path = bin_dir + "maltparser-1.9.1" + os.sep
marmot_path = bin_dir + "Marmot" + os.sep


def log_tasks(opts):
	sys.stderr.write("\nRunning tasks:\n" +"="*20 + "\n")
	if opts.sent not in ["auto","none"]:
		sys.stderr.write("o Splitting sentences based on tag: "+opts.sent+"\n")
	elif opts.sent == "auto":
		sys.stderr.write("o Automatic sentence splitting\n")
	if opts.whitespace:
		sys.stderr.write("o Whitespace tokenization\n")
	if opts.tokenize:
		sys.stderr.write("o Morphological segmentation\n")
	if opts.pos:
		sys.stderr.write("o POS tagging\n")
	if opts.lemma:
		sys.stderr.write("o Lemmatization\n")
	if opts.morph:
		sys.stderr.write("o Morphological analysis\n")
	if opts.dependencies:
		sys.stderr.write("o Dependency parsing\n")
	if opts.entities:
		sys.stderr.write("o Entity recognition\n")
	if opts.coref:
		sys.stderr.write("o Coreference resolution\n")

	sys.stderr.write("\n")


def diagnose_opts(opts):

	if not opts.pos and not opts.morph and not opts.whitespace and not opts.tokenize and not opts.lemma \
		and not opts.dependencies and not opts.entities and not opts.coref:
		if not opts.quiet:
			sys.stderr.write("! You selected no processing options\n")
			sys.stderr.write("! Assuming you want all processing steps\n")
		opts.whitespace = True
		opts.tokenize = True
		opts.pos = True
		opts.morph = True
		opts.lemma = True
		opts.dependencies = True
		opts.entities = True
		opts.coref = True

	added = []
	trigger = ""
	if opts.dependencies:
		trigger = "depenedencies"
		if not opts.pos:
			added.append("pos")
			opts.pos = True
		if not opts.lemma:
			added.append("lemma")
			opts.lemma = True
		if not opts.morph:
			added.append("morph")
			opts.morph = True
	if len(added)>0:
		sys.stderr.write("! You selected "+trigger+"\n")
		sys.stderr.write("! Turning on options: "+",".join(added) +"\n")

	added = []
	if opts.whitespace:
		trigger = "whitespace tokenization"
		if not opts.tokenize:
			added.append("tokenize")
			opts.tokenize = True
	if len(added)>0:
		sys.stderr.write("! You selected "+trigger+"\n")
		sys.stderr.write("! Turning on options: "+",".join(added) +"\n")
	return opts


def groupify(output,anno):
	groups = ""
	current_group = ""
	for line in output.split("\n"):
		if " "+anno+"=" in line:
			current_group += re.search(anno + r'="([^"]*)"',line).group(1)
		if line.startswith("</") and "_group" in line:
			groups += current_group +"\n"
			current_group = ""

	return groups


def get_bound_group_map(data):

	mapping = {}
	data = data.split("\n")
	# Ignore markup
	data = [u for u in data if not (u.startswith("<") and u.endswith(">"))]
	counter = 0
	for i, line in enumerate(data):
		super_token = line.replace("|","") if line != "|" else "|"
		segs = line.split("|") if line != "|" else ["|"]
		for j, seg in enumerate(segs):
			if len(segs)>1 and j == 0:
				mapping[counter] = (super_token,len(segs))
				super_token = ""
			counter += 1

	return mapping


def remove_nesting_attr(data, nester, nested, attr="xml:lang"):
	"""
	Removes attribute on nesting element if a nested element includes it

	:param data: SGML input
	:param nester: nesting tag, e.g. "norm"
	:param nested: nested tag, e.g. "morph"
	:param attr: attribute, e.g. "lang"
	:return: cleaned SGML
	"""

	if attr not in data:
		return data
	flagged = []
	in_attr_nester = False
	last_nester = -1
	lines = data.split("\n")
	for i, line in enumerate(lines):
		if nester + "=" in line and attr+"=" in line:
			in_attr_nester = True
			last_nester = i
		if "</" + nester + ">" in line:
			in_attr_nester = False
		if nested in line and attr+"=" in line and in_attr_nester and last_nester > -1:
			flagged.append(last_nester)
			in_attr_nester = False
	for i in flagged:
		lines[i] = re.sub(' '+attr+'="[^"]+"','',lines[i])
	return "\n".join(lines)


def tok_from_norm(data):
	"""
	Takes TT-SGML, extracts norm attribute, and replaces existing tokens with norm values while retaining SGML tags.
	Used to feed parser norms while retaining SGML sentence separators.

	:param data: TTSGML with <norm norm=...> and raw tokens to replace
	:return: TTSGML with tags preserved and tokens replace by norm attribute values
	"""

	outdata = []
	norm = ""
	for line in data.replace("\r","").split("\n"):
		if line.startswith("<"):
			m = re.search(r'norm="([^"]*)"',line)
			if m is not None:
				norm = m.group(1)
			outdata.append(line)
		else:
			if norm != "":
				outdata.append(norm)
				norm=""
	return "\n".join(outdata) + "\n"


def read_attributes(input,attribute_name):
	out_stream =""
	for line in input.split('\n'):
		if attribute_name + '="' in line:
			m = re.search(attribute_name+r'="([^"]*)"',line)
			if m is None:
				print("ERR: cant find " + attribute_name + " in line: " + line)
				attribute_value = ""
			else:
				attribute_value = m.group(1)
			if len(attribute_value)==0:
				attribute_value = "_warn:empty_"+attribute_name+"_"
			out_stream += attribute_value +"\n"
	return out_stream


def merge_into_tag(tag_to_kill, tag_to_merge_into,stream):
	vals = []
	cleaned_stream = ""
	for line in stream.split("\n"):
		if " "+tag_to_kill + "=" in line:
			val = re.search(" " + tag_to_kill+'="([^"]*)"',line).group(1)
			vals.append(val)
		elif "</" + tag_to_kill + ">" in line:
			pass
		else:
			cleaned_stream += line + "\n"
	injected = inject(tag_to_kill,"\n".join(vals).strip(),tag_to_merge_into,cleaned_stream)
	return injected


def get_col(data, colnum):
	if not isinstance(data,list):
		data = data.split("\n")

	splits = [row.split("\t") for row in data if "\t" in row]
	return [r[colnum] for r in splits]


def exec_via_temp(input_text, command_params, workdir="", outfile=False):
	temp = tempfile.NamedTemporaryFile(delete=False)
	if outfile:
		temp2 = tempfile.NamedTemporaryFile(delete=False)
	output = ""
	try:
		temp.write(input_text.encode("utf8"))
		temp.close()

		if outfile:
			command_params = [x if 'tempfilename2' not in x else x.replace("tempfilename2",temp2.name) for x in command_params]
		command_params = [x if 'tempfilename' not in x else x.replace("tempfilename",temp.name) for x in command_params]
		if workdir == "":
			proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE)
			(stdout, stderr) = proc.communicate()
		else:
			proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
			(stdout, stderr) = proc.communicate()
		if outfile:
			if PY3:
				output = io.open(temp2.name,encoding="utf8").read()
			else:
				output = open(temp2.name).read()
			temp2.close()
			os.remove(temp2.name)
		else:
			output = stdout
		#print(stderr)
		proc.terminate()
	except Exception as e:
		print(e)
	finally:
		os.remove(temp.name)
		return output


def exec_via_temp_old(input_text, command_params, workdir=""):
	temp = tempfile.NamedTemporaryFile(delete=False)
	exec_out = ""
	try:
		if PY3:
			temp.write(input_text.encode("utf8"))
		else:
			temp.write(input_text.encode("utf8"))
		temp.close()

		command_params = [x if x != 'tempfilename' else temp.name for x in command_params]
		if workdir == "":
			proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE)
			(stdout, stderr) = proc.communicate()
		else:
			proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
			(stdout, stderr) = proc.communicate()

		exec_out = stdout
	except Exception as e:
		print(e)
	finally:
		os.remove(temp.name)
		#if PY3:
		exec_out = exec_out.decode("utf8")
		return exec_out


def get_origs(data):
	origs = []
	current = ""
	for line in data.split("\n"):
		if "</norm>" in line:
			origs.append(current)
			current = ""
		if not line.startswith("<"):  # Token line
			current += line

	return "\n".join(origs)


def inject(attribute_name, contents, at_attribute,into_stream,replace=True):
	insertions = contents.split('\n')
	injected = ""
	i=0
	for line in into_stream.split("\n"):
		if at_attribute + "=" in line:
			if i >= len(insertions):
				raise Exception("Error out of bounds at element " + str(i) + " in document beginning " + into_stream[:1000])
			if len(insertions[i])>0:
				if at_attribute == attribute_name:  # Replace old value of attribute with new one
					line = re.sub(attribute_name+'="[^"]+"',attribute_name+'="'+insertions[i]+'"',line)
				else:  # Place before specific at_attribute
					if replace or " " + attribute_name + "=" not in line:
						line = re.sub(at_attribute+"=",attribute_name+'="'+insertions[i]+'" '+at_attribute+"=",line)
			i += 1
		injected += line + "\n"
	return injected


def extract_conll(conll_string):
	conll_string = conll_string.replace("\r","").strip()
	sentences = conll_string.split("\n\n")
	ids = ""
	funcs = ""
	parents = ""
	id_counter = 0
	offset = 0
	for sentence in sentences:
		tokens = sentence.split("\n")
		for token in tokens:
			if "\t" in token:
				id_counter +=1
				ids += "u"+ str(id_counter) + "\n"
				cols = token.split("\t")
				funcs += cols[7].replace("ROOT","root") +"\n"
				if cols[6] == "0":
					parents += "#u0\n"
				else:
					parents += "#u" + str(int(cols[6])+offset)+"\n"
		offset = id_counter
	return ids, funcs, parents


def inject_tags(in_sgml,insertion_specs,around_tag="norm",inserted_tag="mwe"):
	"""

	:param in_sgml: input SGML stream including tags to surround with new tags
	:param insertion_specs: list of triples (start, end, value)
	:param around_tag: tag of span to surround by insertion
	:return: modified SGML stream
	"""
	if len(insertion_specs) == 0:
		return in_sgml

	counter = -1
	next_insert = insertion_specs[0]
	outlines = []
	for line in in_sgml.split("\n"):
		if line.startswith("<" + around_tag + " "):
			counter += 1
			if next_insert[0] == counter:  # beginning of a span
				outlines.append("<" + inserted_tag + " " + inserted_tag + '="' + next_insert[2] + '">')
		outlines.append(line)
		if line.startswith("</" + around_tag + ">"):
			if next_insert[1] == counter:  # end of a span
				outlines.append("</" + inserted_tag + ">")

	return "\n".join(outlines)


def check_requirements():
	marmot_OK = True
	malt_OK = True
	models_OK = True
	marmot = marmot_path + "marmot.jar"
	if not os.path.exists(marmot):
		sys.stderr.write("! Marmot not found at ./bin/\n")
		marmot_OK = False
	if not os.path.exists(parser_path+"maltparser-1.9.1.jar"):
		sys.stderr.write("! Malt Parser 1.9.1 not found at ./bin/\n")
		malt_OK = False
	model_files = ["heb.sm" + str(sys.version_info[0]), "heb.xrm", "heb.mco", "heb.marmot", "heb.lemming"]
	for model_file in model_files:
		if not os.path.exists(model_dir + model_file):
			sys.stderr.write("! Model file " + model_file + " missing in ./models/\n")
			models_OK = False

	return marmot_OK, malt_OK, models_OK


def download_requirements(marmot_ok=True, malt_ok=True, models_ok=True):
	import requests, zipfile, shutil, tarfile
	if not PY3:
		import StringIO
	urls = []
	if not malt_ok:
		urls.append("http://maltparser.org/dist/maltparser-1.9.1.tar.gz")
	if not marmot_ok:
		if not os.path.exists(bin_dir + "Marmot"):
			os.makedirs(bin_dir + "Marmot")
		marmot_base_url = "http://cistern.cis.lmu.de/marmot/bin/CURRENT/"
		marmot_current = requests.get(marmot_base_url).text
		files = re.findall(r'href="((?:marmot|trove)[^"]+jar)"',marmot_current)
		marmot_file = ""
		trove_file = ""
		for f in files:
			if f.startswith("marmot"):
				marmot_file = f
			elif f.startswith("trove"):
				trove_file = f
		urls.append(marmot_base_url + marmot_file)
		urls.append(marmot_base_url + trove_file)
	if not models_ok:
		models_base = "http://corpling.uis.georgetown.edu/amir/download/heb_models/"
		urls.append(models_base + "heb.sm" + str(sys.version_info[0]))
		urls.append(models_base + "heb.mco")
		urls.append(models_base + "heb.xrm")
		urls.append(models_base + "heb.lemming")
		urls.append(models_base + "heb.marmot")
	for u in urls:
		sys.stderr.write("o Downloading from " + str(u) + "\n")
		if "corpling" in u:
			base_name = u[u.rfind("/") + 1:]
			urlretrieve(u,model_dir + base_name)
		else:
			r = requests.get(u, stream=True)
			if PY3:
				file_contents = io.BytesIO(r.content)
			else:
				file_contents = StringIO.StringIO(r.content)
			if u.endswith("gz"):
				z = tarfile.open(fileobj=file_contents, mode="r:gz")
				z.extractall(path=bin_dir)
			elif u.endswith("jar"):
				if "trove" in u:
					with open(bin_dir + "Marmot" + os.sep + "trove.jar", 'wb') as f:
						f.write(r.content)
				elif "marmot" in u:
					with open(bin_dir + "Marmot" + os.sep + "marmot.jar", 'wb') as f:
						f.write(r.content)
	sys.stderr.write("\n")
	# Copy java dependency model files to tool working dirs
	shutil.copyfile(model_dir+"heb.mco",bin_dir+"maltparser-1.9.1" + os.sep + "heb.mco")
	shutil.copyfile(model_dir+"heb.marmot",bin_dir+"Marmot" + os.sep + "heb.marmot")
	shutil.copyfile(model_dir+"heb.lemming",bin_dir+"Marmot" + os.sep + "heb.lemming")


def nlp(input_data, do_whitespace=True, do_tok=True, do_tag=True, do_lemma=True, do_parse=True, do_entity=True,
		out_mode="conllu", sent_tag=None, preloaded=None):

	data = input_data.replace("\t","")
	data = data.replace("\r","")

	if preloaded is not None:
		rf_tok, xrenner = preloaded
	else:
		rf_tok = RFTokenizer(model=model_dir + "heb.sm" + str(sys.version_info[0]))
		xrenner = Xrenner(model=model_dir + "heb.xrm")

	if do_whitespace:
		data = whitespace_tokenize(data, abbr=data_dir + "heb_abbr.tab",add_sents=sent_tag=="auto")

	if do_tok:
		tokenized = rf_tok.rf_tokenize(data.strip().split("\n"))
		tokenized = "\n".join(tokenized)
	else:
		# Assume data is already one token per line
		tokenized = data

	bound_group_map = get_bound_group_map(tokenized) if out_mode == "conllu" else None

	if sent_tag == "auto":
		tokenized = toks_to_sents(tokenized)
		sent_tag = "s"

	if out_mode == "pipes":
		return tokenized
	else:
		tokenized = tokenized.split("\n")
		retokenized = []
		for line in tokenized:
			if line == "|":
				retokenized.append(line)
			else:
				retokenized.append("\n".join(line.split("|")))
		tokenized = "\n".join(retokenized)

	if do_tag:
		if platform.system() == "Windows":
			tag = ["java","-Dfile.encoding=UTF-8","-Xmx2g","-cp","marmot.jar;trove.jar","marmot.morph.cmd.Annotator","-model-file","heb.marmot","-lemmatizer-file","heb.lemming","-test-file","form-index=0,tempfilename","-pred-file","tempfilename2"]
		else:
			tag = ["java","-Dfile.encoding=UTF-8","-Xmx2g","-cp","marmot.jar:trove.jar","marmot.morph.cmd.Annotator","-model-file","heb.marmot","-lemmatizer-file","heb.lemming","-test-file","form-index=0,tempfilename","-pred-file","tempfilename2"]
		no_sent = re.sub(r'</?s>\n?','',tokenized).strip()
		morphed = exec_via_temp(no_sent, tag, workdir=marmot_path, outfile=True)
		morphed = morphed.strip().split("\n")
		morphs = get_col(morphed,7)
		lemmas = get_col(morphed,3)

		tagged = inject_col(morphed,tokenized,5)
		if do_lemma:
			lemmatized = inject_col(lemmas,tagged,-1)
		else:
			lemmatized = tagged
		morphed = inject_col(morphs,lemmatized,-1)

		if not do_parse:
			if out_mode == "conllu":
				conllized = conllize(morphed, tag="PUNCT", element=sent_tag, no_zero=True, super_mapping=bound_group_map)
				return conllized
			else:
				if not PY3:
					morphed = morphed.decode("utf8")
				return morphed

	else:
		if out_mode == "conllu":
			conllized = conllize(tokenized, tag="PUNCT", element=sent_tag, no_zero=True, super_mapping=bound_group_map)
			return conllized
		else:
			return tokenized

	if do_parse:
		conllized = conllize(morphed, tag="PUNCT", element=sent_tag, no_zero=True, super_mapping=bound_group_map)
		parse = ['java','-mx1g','-jar',"maltparser-1.9.1.jar",'-c','heb','-i','tempfilename','-m','parse']
		parsed = exec_via_temp(conllized,parse,parser_path)

		if do_entity:
			xrenner.docname = "_"
			if PY3:
				parsed = parsed.decode("utf8")
			ents = xrenner.analyze(parsed,"conll_sent")
			ents = get_col(ents, -1)
			entified = inject_col(ents, parsed, col=-1, into_col=9, skip_supertoks=True)
			if PY3:
				return entified
			else:
				return entified.decode("utf8")
		else:
			return parsed.decode("utf8")
	else:
		if out_mode == "conllu":
			conllized = conllize(tagged, tag="PUNCT", element=sent_tag, no_zero=True, super_mapping=bound_group_map)
			return conllized
		else:
			return tagged


def run_hebpipe():


	if sys.version_info[0] == 2 and sys.version_info[1] < 7:
		sys.stderr.write("Python versions below 2.7 are not supported.\n")
		sys.stderr.write("Your Python version:\n")
		sys.stderr.write(".".join([str(v) for v in sys.version_info[:3]]) + "\n")
		sys.exit(0)

	from argparse import ArgumentParser, RawDescriptionHelpFormatter

	parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter)
	parser.prog = "HebPipe - NLP Pipeline for Hebrew"
	parser.usage = "python heb_pipe.py [OPTIONS] files"
	parser.epilog = """Example usage:
--------------
Whitespace tokenize, tokenize morphemes, add pos, lemma, morph, dep parse with automatic sentence splitting, 
entity recognition and coref for one text file, output in default conllu format:
> python heb_pipe.py -wtplmdec example_in.txt        

OR specify no processing options (automatically assumes you want all steps)
> python heb_pipe.py example_in.txt        

Just tokenize a file using pipes:
> python heb_pipe.py -wt -o pipes example_in.txt     

Pos tag, lemmatize, add morphology and parse a pre-tokenized file, splitting sentences by existing <sent> tags:
> python heb_pipe.py -plmd -s sent example_in.txt  

Add full analyses to a whole directory of *.txt files, output to a specified directory:    
> python heb_pipe.py -wtplmdec --dirout /home/heb/out/ *.txt

Parse a tagged TT SGML file into CoNLL tabular format for treebanking, use existing tag <sent> to recognize sentence borders:
> python heb_pipe.py -d -s sent example_in.tt
"""
	parser.add_argument("files", help="File name or pattern of files to process (e.g. *.txt)")

	g1 = parser.add_argument_group("standard module options")
	g1.add_argument("-w","--whitespace", action="store_true", help='Perform white-space based tokenization of large word forms')
	g1.add_argument("-t","--tokenize", action="store_true", help='Tokenize large word forms into smaller morphological segments')
	g1.add_argument("-p","--pos", action="store_true", help='Do POS tagging')
	g1.add_argument("-l","--lemma", action="store_true", help='Do lemmatization')
	g1.add_argument("-m","--morph", action="store_true", help='Do morphological tagging')
	g1.add_argument("-d","--dependencies", action="store_true", help='Parse with dependency parser')
	g1.add_argument("-e","--entities", action="store_true", help='Add entity spans and types')
	g1.add_argument("-c","--coref", action="store_true", help='Add coreference annotations')
	g1.add_argument("-s","--sent", action="store", default="auto", choices=["auto","none"], help='XML tag to split sentences, e.g. sent for <sent ..> or none for no splitting (otherwise automatic sentence splitting)')
	g1.add_argument("-o","--out", action="store", choices=["pipes","conllu","sgml"], default="conllu", help='Output CoNLL format, SGML or just tokenize with pipes')

	g2 = parser.add_argument_group("less common options")
	g2.add_argument("-q","--quiet", action="store_true", help='Suppress verbose messages')
	g2.add_argument("-x","--extension", action="store", default='conllu', help='Extension for output files (default: .conllu)')
	g2.add_argument("--dirout", action="store", default=".", help='Optional output directory (default: this dir)')
	g2.add_argument("--version", action="store_true", help='Print version number and quit')

	if "--version" in sys.argv:
		sys.stdout.write("HebPipe V" + __version__)
		sys.exit(1)

	opts = parser.parse_args()
	opts = diagnose_opts(opts)
	dotok = opts.tokenize

	if not opts.quiet:
		try:
			from .lib import timing
		except ImportError:  # direct script usage
			from lib import timing

	files = glob(opts.files)

	if not opts.quiet:
		log_tasks(opts)

	# Check if models, Marmot and Malt Parser are available
	if opts.pos or opts.lemma or opts.morph or opts.dependencies or opts.tokenize or opts.entities:
		marmot_OK, malt_OK, models_OK = check_requirements()
		if ((opts.pos or opts.lemma or opts.morph) and not marmot_OK) or (opts.dependencies and not malt_OK) or not models_OK:
			sys.stderr.write("! You are missing required software:\n")
			if (opts.pos or opts.lemma or opts.morph) and not marmot_OK:
				sys.stderr.write(" - Tagging, lemmatization and morphological analysis require Marmot\n")
			if opts.dependencies and not malt_OK:
				sys.stderr.write(" - Parsing is specified but Malt Parser 1.9.1 is not installed\n")
			if not models_OK:
				sys.stderr.write(" - Model files in models/ are missing\n")
			response = inp("Attempt to download missing files? [Y/N]\n")
			if response.upper().strip() == "Y":
				download_requirements(marmot_OK,malt_OK,models_OK)
			else:
				sys.stderr.write("Aborting\n")
				sys.exit(0)

	if dotok:  # Pre-load stacked tokenizer for entire batch
		rf_tok = RFTokenizer(model=model_dir + "heb.sm" + str(sys.version_info[0]))
	else:
		rf_tok = None
	if opts.entities:  # Pre-load stacked tokenizer for entire batch
		xrenner = Xrenner(model=model_dir + "heb.xrm")
	else:
		xrenner = None

	for infile in files:
		base = os.path.basename(infile)
		if infile.endswith("." + opts.extension):
			outfile = base.replace("." + opts.extension,".out." + opts.extension)
		elif len(infile) > 4 and infile[-4] == ".":
			outfile = base[:-4] + "." + opts.extension
		else:
			outfile = base + "." + opts.extension

		if not opts.quiet:
			sys.stderr.write("Processing " + base + "\n")

		input_text = io.open(infile,encoding="utf8").read()


		processed = nlp(input_text, do_whitespace=opts.whitespace, do_tok=dotok, do_tag=opts.pos, do_lemma=opts.lemma,
							   do_parse=opts.dependencies, out_mode=opts.out,
							   sent_tag=opts.sent, preloaded=(rf_tok,xrenner))

		if len(files) > 1:
			with io.open(opts.dirout + os.sep + outfile, 'w', encoding="utf8", newline="\n") as f:
				if not PY3:
					processed = unicode(processed)
				f.write((processed.strip() + "\n"))
		else:  # Single file, print to stdout
			if PY3:
				sys.stdout.buffer.write(processed.encode("utf8"))
			else:
				print(processed.encode("utf8"))

	fileword = " files\n\n" if len(files) > 1 else " file\n\n"
	sys.stderr.write("\nFinished processing " + str(len(files)) + fileword)

if __name__ == "__main__":
	run_hebpipe()


#!/usr/bin/python
# -*- coding: utf-8 -*-

import io, sys
from argparse import ArgumentParser

#this script takes a text file with tab delimited columns
#and adds every column except the first line-wise to the target file.
#Both files should ideally have the same number of lines.
#The script will terminate once either file ends.
# For example:
# Appended file:
# Computer	Nom.Sg.Masc
#
# Target file:
# Computer	NN	Computer 
#
# becomes
#
# Computer	NN	Computer	Nom.Sg.Masc
#
# usage:
# append_column.py [OPTIONS] <APPEND_COLS_FILE> <TARGET_FILE>
#
# Options and arguments:
#
# -h           Print this message and quit
#
#
# example:
# append_column.py morph.tab tagged.tab > merged.tab
# etc.

PY3 = sys.version_info[0] == 3


def inject_col(source_lines, target_lines, col=-1, into_col=None, skip_supertoks=False):

	output = []
	counter = -1
	target_line = ""

	if not PY3:
		if isinstance(target_lines,unicode):
			target_lines = str(target_lines.encode("utf8"))

	if not isinstance(source_lines,list):
		source_lines = source_lines.split("\n")
	if not isinstance(target_lines,list):
		target_lines = target_lines.split("\n")

	for i, source_line in enumerate(source_lines):
		while len(target_line) == 0:
			counter +=1
			target_line = target_lines[counter]
			if (target_line.startswith("<") and target_line.endswith(">")) or len(target_line) == 0:
				output.append(target_line)
				target_line = ""
			else:
				target_cols = target_line.split("\t")
				if "-" in target_cols[0] and skip_supertoks:
					output.append(target_line)
					target_line = ""
		source_cols = source_line.split("\t")
		to_inject = source_cols[col]
		target_cols = target_line.split("\t")
		if into_col is None:
			target_cols.append(to_inject)
		else:
			target_cols[into_col] = to_inject
		output.append("\t".join(target_cols))
		target_line=""

	return "\n".join(output)


if __name__ == "__main__":
	p = ArgumentParser()
	p.add_argument("source")
	p.add_argument("target")
	p.add_argument("-c","--col",action="store",default="-1")

	opts = p.parse_args()
	source_lines = io.open(opts.file1,encoding="utf8").read().replace("\r","").split("\n")
	target_lines = io.open(opts.file1,encoding="utf8").read().replace("\r","").split("\n")
	merged = inject_col(source_lines,target_lines,opts.col)
	print(merged)

